/*
	$Id: Help.java 6881 2015-04-21 07:59:42Z nicklas $

	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/

package net.sf.basedb.clients.web.taglib;


import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;

/**
  This tag is used to make it easier to add help features on a BASE web page.
  <p>
  <b>Example:</b>
  <pre>
  	&lt;%@ taglib prefix="base" uri="/WEB-INF/base.tld" %&gt;
  	&lt;base:help help_id="required_helpexternalID" /&gt;
  </pre>
  
  <p>
  <b>Syntax:</b>
  <pre>
  &lt;base:help
  	helpid=...
  	tabcontrol=...
  	clazz=...
  	subclass=...
  	style=...
  	image=...
  	tooltip=...
  	visible=...
  	tabindex=...
  &gt;
  </pre>

  	<table border=1 cellspacing=0 cellpadding=2>
 	<caption>Tag attributes</caption>
 	<tr>
	  	<td><b>Attribute</b></td>
	  	<td><b>Default value</b></td>
	  	<td><b>Required</b></td>
	  	<td><b>Description</b></td>
  	</tr>
  	<tr>
  		<td>helpid</td>
  		<td></td>
  		<td>no</td>
  		<td>
  			The keyword of the help topic to display. Each help topic has a one-word key
  			to be identified with. If not specified then the <code>tabcontrol</code>
  			attribute must be specified. 			 
  		</td>
  	</tr>  	
  	<tr>
  		<td>tabcontrol</td>
  		<td></td>
  		<td>no</td>
  		<td>
  			The ID of the {@link net.sf.basedb.clients.web.taglib.tab.TabControl} to get the 
  			help id from. Each {@link net.sf.basedb.clients.web.taglib.tab.Tab} in a tabcontrol 
  			should specify it's own <code>helpid</code> attribute.
  		</td>
  	</tr>  	
  	<tr>
  		<td>clazz</td>
  		<td>helpicon</td>
  		<td>no</td>
  		<td>
  			The value if this attribute goes directly into the standard HTML 
			<code>class</code> attribute. This allows you to apply different styles
			to different helps.  		
  		</td>
  	</tr>
	<tr>
		<td>subclass</td>
		<td>-</td>
		<td>no</td>
		<td>
			Additional classes to put in the HTML class attribute. Doesn't
			replace the main class attribute.
		</td>
	</tr>
  	<tr>
  		<td>style</td>
  		<td>-</td>
  		<td>no</td>
  		<td>
  			The value if this attribute goes directly into the standard HTML 
			<code>style</code> attribute. This allows you to apply different styles
			to different icons.
  		</td>
  	</tr>
  	<tr>
  		<td>image</td>
  		<td>-</td>
  		<td>no</td>
  		<td>
  			The name of the image to display. The image must be in the <code>/images</code>
			directory. Only include the filename, not the path.
			This taglib will find the correct path from the current jsp page.
		</td>
	</tr>
  	<tr>
  		<td>tooltip</td>
  		<td>-</td>
  		<td>no</td>
  		<td>
  			A short explanation that will show up as a tooltip.
  		</td>
  	</tr>
  	<tr>
  		<td>visible</td>
  		<td>true</td>
  		<td>no</td>
  		<td>
  			If the help icon should be visible or not.
  		</td>
  	</tr>
 	<tr>
		<td>tabindex</td>
		<td>0</td>
		<td>no</td>
		<td>
			Tab index value to control the focus order. Use -1 to disable tabbing to the icon.
		</td>
	</tr>
  	</table>
  	
  
  @author martin
  @version 2.0
*/
public class Help 
	extends StylableTag
{
	/**
		
	*/
	private static final long serialVersionUID = -2754197259139443681L;

	/**
	 	The parent &lt;base:page&gt; tag.	 	
	*/
	private Page page;
		
	/**
	 	The keyword of the helptopic
	*/
	private String helpid = null;
	
	/**
		The tabcontrol ID to get the help id from.
	*/
	private String tabcontrol = null;
	
	
	/**
	  	The path to the help icon
	*/
	private String image = "help.png";
	
	/**
	 	Optional tooltip.
	*/
	private String tooltip = null;
	
	/**
	 * 	If the help icon should be visible or not.
	 */
	private boolean visible = true;
	
	private int tabIndex;
	
	public Help()
	{
		super("helpicon");
	}
	
	/*
	 	Taglib initialization methods
	*/
	public void setHelpid(String helpid)
	{
		this.helpid = helpid;
	}
	public String getHelpid()
	{
		return this.helpid;
	}
	
	public void setTabcontrol(String tabcontrol)
	{
		this.tabcontrol = tabcontrol;
	}
	public String getTabcontrol()
	{
		return this.tabcontrol;
	}

	public void setImage(String image)
	{
		this.image = image;		
	}
	public String getImage()
	{
		return this.image;
	}
	
	public void setTooltip(String tooltip)
	{
		this.tooltip = tooltip;
	}
	public String getTooltip()
	{
		return this.tooltip;
	}
	
	public void setVisible(boolean visible)
	{
		this.visible = visible;
	}
	public boolean isVisible()
	{
		return this.visible;
	}
	
	/**
		@since 3.2
	*/
	public void setTabindex(int tabIndex)
	{
		this.tabIndex = tabIndex;
	}
	
	/**
		@since 3.2
	*/
	public int getTabindex()
	{
		return tabIndex;
	}
	
	@Override
	public int doStartTag() 
		throws JspException
	{
		page = (Page)findAncestorWithClass(this, Page.class);
		if (page == null) throw new JspTagException("Tag <base:help > must be inside a <base:page> tag");
		if (!isVisible()) return SKIP_BODY;
		
		StringBuilder sb = new StringBuilder();
		sb.append("<img");
		addIdAndStyles(sb, "auto-init");
		sb.append(" src=\"").append(page.getRoot()).append("images/").append(getImage()).append("\"");
		sb.append(" data-auto-init=\"help\"");
		if (getHelpid() != null)
		{
			sb.append(" data-help-id=\"").append(getHelpid()).append("\"");
		}
		if (getTabcontrol() != null)
		{
			sb.append(" data-tabcontrol-id=\"").append(getTabcontrol()).append("\"");
		}
		if (getTabindex() >= 0) 
		{
			sb.append(" tabindex=\"").append(getTabindex()).append("\"");
		}
		sb.append(" alt=\"?\"");
		sb.append(">");
		try
		{
			pageContext.getOut().print(sb.toString());
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}		
		return EVAL_BODY_INCLUDE;		
	}
}
