/*
	$Id: Navigator.java 6881 2015-04-21 07:59:42Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.taglib.table;

import java.util.TreeSet;

import javax.servlet.jsp.JspException;
import javax.servlet.jsp.JspTagException;

import net.sf.basedb.clients.web.taglib.StylableTag;

/**
	This tag is used to display a navigator for a pageable table
	of items. A pageable table is a table where only a limited (for
	example 30) items are show at once. The navigator makes it
	possible to step forwards and backwards among the pages.
	This tag must appear inside a &lt;tbl:table&gt; tag.
	<p>
	<b>Syntax:</b>
	<pre>
   &lt;tbl:navigator
      id=...
      clazz=...
      style=...
      page=...
      rowsperpage=...
      totalrows=...
      visible=true|false
   &gt;
</pre>

	<table border=1 cellspacing=0 cellpadding=2>
	<caption>Tag attributes</caption>
	<tr>
		<td><b>Attribute</b></td>
		<td><b>Default value</b></td>
		<td><b>Required</b></td>
		<td><b>Description</b></td>
	</tr>
	<tr>
		<td>id</td>
		<td>-</td>
		<td>no</td>
		<td>
			The ID of the navigator. The value of this attribute
			goes directly into the standard <code>id</code> attribute.
			This means that if you need a reference to the object in your own
			JavaScript code, you may for example use the following code:
			<p>
			<code>
			navigator = document.getElementById('&lt;id&gt;');
			</code>
		</td>
	</tr>
	<tr>
		<td>clazz</td>
		<td>panel</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>class</code> attribute. This allows you to apply different styles
			to different tables.
		</td>
	</tr>
	<tr>
		<td>style</td>
		<td>-</td>
		<td>no</td>
		<td>
			The value if this attribute goes directly into the standard HTML 
			<code>style</code> attribute. This allows you to apply different styles
			to different tables.
		</td>
	</tr>
	<tr>
		<td>page</td>
		<td>0</td>
		<td>no</td>
		<td>The current page, starting from 0.
		</td>
	</tr>
	<tr>
		<td>rowsperpage</td>
		<td>30</td>
		<td>no</td>
		<td>The number of rows on a single page.</td>
	</tr>
	<tr>
		<td>totalrows</td>
		<td>-</td>
		<td>yes</td>
		<td>The total number of rows.</td>
	</tr>
	<tr>
		<td>visible</td>
		<td>true</td>
		<td>no</td>
		<td>
			If the navigator bar should be visible or not.
		</td>
	</tr>
	</table>
	<p>
	This tag has no subtags.

	@author Nicklas
	@version 2.0
*/
public class Navigator
	extends StylableTag
{

	/**
		
	*/
	private static final long serialVersionUID = 3287798191943288674L;

	public Navigator()
	{
		super("navigator");
	}
	
	/**
		The parent &lt;tbl:table&gt; tag.
	*/
	private Table table;

	/**
		The current page.
	*/
	private int currentPage = 0;
	
	/**
		The number of rows to display on a single page.
	*/
	private int rowsPerPage = 30;
	
	/**
		The total number of rows.
	*/
	private int totalRows = 0;

	/**
		If the navigator bar should be visible or not.
	*/
	private boolean visible = true;

	private boolean locked = false;

	/*
		Taglib initialization methods
	*/
	public void setPage(String page)
	{
		setPage(Integer.parseInt(page));
	}
	public void setPage(int page)
	{
		if (page < 0) page = 0;
		this.currentPage = page;
	}
	public int getPage()
	{
		return currentPage;
	}

	public void setRowsperpage(String rowsPerPage)
	{
		setRowsperpage(Integer.parseInt(rowsPerPage));
	}
	public void setRowsperpage(int rowsPerPage)
	{
		if (rowsPerPage == 0) rowsPerPage = 30;
		this.rowsPerPage = rowsPerPage;
	}
	public int getRowsperpage()
	{
		return rowsPerPage;
	}

	public void setTotalrows(String totalRows)
	{
		setTotalrows(Integer.parseInt(totalRows));
	}
	public void setTotalrows(int totalRows)
	{
		this.totalRows = totalRows;
	}
	public void setTotalrows(long totalRows)
	{
		this.totalRows = (int)totalRows;
	}
	public int getTotalrows()
	{
		return totalRows;
	}

	public void setVisible(boolean visible)
	{
		this.visible = visible;
	}
	public boolean isVisible()
	{
		return visible;
	}
	
	public void setLocked(boolean locked)
	{
		this.locked = locked;
	}
	
	public boolean isLocked()
	{
		return locked;
	}

	@Override
	public int doEndTag()
		throws JspException
	{
		table = (Table)findAncestorWithClass(this, Table.class);
		if (!isVisible()) return SKIP_BODY;
		
		if (!isLocked())
		{
			table.addHidden("page", Integer.toString(currentPage));
		}

		String root = table.getPage().getRoot();

		StringBuilder sb = new StringBuilder();
		sb.append("<div");
		addIdAndStyles(sb, "table-navigator");
		sb.append(" data-table-id=\"").append(table.getId()).append("\"");
		addDynamicAttributes(sb);
		sb.append(">");
		sb.append("<span class=\"flowcontrol\">");
		int numPages = (int)Math.ceil((double)totalRows / (double)rowsPerPage);
		if (currentPage > 0)
		{
			sb.append("<img class=\"link\" data-table-page=\"0\" title=\"Goto first page\" src=\"").append(root).append("images/first.png\" alt=\"|◄\">");
			sb.append("<img class=\"link\" data-table-page=\"").append(currentPage-1).append("\" title=\"Goto previous page\" src=\"").append(root).append("images/previous.png\" alt=\"◄\">");
		}
		else
		{
			sb.append("<img src=\"").append(root).append("images/first.png\" class=\"disabled\" alt=\"\">");
			sb.append("<img src=\"").append(root).append("images/previous.png\" class=\"disabled\" alt=\"\">");
		}
		sb.append("</span><span class=\"flowcontrol\">");
		if ((currentPage+1)*rowsPerPage < totalRows )
		{
			sb.append("<img class=\"link\" data-table-page=\"").append(currentPage+1).append("\" title=\"Goto next page\" src=\"").append(root).append("images/next.png\" alt=\"►\">");
			sb.append("<img class=\"link\" data-table-page=\"").append(numPages-1).append("\" title=\"Goto last page\" src=\"").append(root).append("images/last.png\" alt=\"►|\">");
		}
		else
		{
			sb.append("<img src=\"").append(root).append("images/next.png\" class=\"disabled\" alt=\"\">");
			sb.append("<img src=\"").append(root).append("images/last.png\" class=\"disabled\" alt=\"\">");
		}
		sb.append("</span>");

		int offset = 1;
		int tmpPage = currentPage;
		TreeSet<Integer> pages = new TreeSet<Integer>();
		for (int i = 1; i < numPages; i += offset)
		{
			int highPage = tmpPage + i;
			int lowPage = tmpPage - i;
			if(highPage < numPages)
			{
				pages.add(highPage);
			}
			if(lowPage > 0 && lowPage < numPages)
			{
				pages.add(lowPage);
			}
			if(i % (offset * 5) == 0)
			{
				offset *= 10;
				tmpPage = (currentPage / offset) * offset - 1;
				i = 0;
			}
		}
		pages.add(0);
		pages.add(currentPage);
		if (numPages > 0) pages.add(numPages - 1);
		sb.append("<span class=\"pages\">");
		for (int page : pages)
		{
			if (page == currentPage)
			{
				sb.append("<span class=\"currentpage\">").append(page+1).append("</span>");
			}
			else
			{
				sb.append("<span class=\"link\" data-table-page=\"").append(page).append("\" title=\"Goto page ").append(page+1).append("\">").append(page+1).append("</span>");
			}
		}
		sb.append("</span>");
		sb.append("<span class=\"rows-per-page\">");
		sb.append("(").append(totalRows).append(totalRows == 1 ? " hit" : " hits");
		sb.append(", <input name=\"rowsperpage\" class=\"text\" type=\"text\" value=\"").append(rowsPerPage).append("\"");
		sb.append(" maxlength=\"3\"> per page)");
		sb.append("</span></div>");
		try
		{
			pageContext.getOut().print(sb.toString());
		}
		catch (Exception ex)
		{
			throw new JspTagException(ex.getMessage());
		}
		return SKIP_BODY;
	}

}
