/*
	$Id: HTML.java 7243 2016-11-28 14:30:47Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.util;

import java.util.regex.*;

import net.sf.basedb.util.EmailUtil;

/**
	This class contains a set of static methods that may be useful 
	in a web application for handling text/HTML strings.
	
	@author Nicklas
	@version 2.0
*/
public class HTML
{

	/**
		This flag is used in {@link #scanForLinks(String,int)}
		when you want to create links for URL:s.
	*/
	public static final int LINK_URL = 1;
	/**
		This flag is used in {@link #scanForLinks(String,int)}
		when you want to create links for email addresses.
	*/
	public static final int LINK_EMAIL = 2;
	
	/**
		This flag is used in {@link #scanForLinks(String,int)}
		when you want to create links for images with a &lt;img&gt; tag instead of
		an &lt;a&gt; tag.
	*/
	public static final int SMART_IMAGES = 4;

	/**
		This pattern can be used to find email addresses. The pattern will check that
		an @ symbol is present and that it is preceded with at least one character and
		followed by at least one subdomain and one topdomain. The pattern will allow
		any unicode letters, digits, underscore and hyphen in the address. It will
		not check that the domain or email address actually exists.
	*/
	public static final Pattern EMAIL_REGEXP = EmailUtil.EMAIL_REGEXP;

	/**
		This pattern can be used to find URL:s. It will look for sequences starting
		with <code>http://</code>, <code>https://</code>, <code>ftp://</code> or 
		<code>www.</code> followed by at least one subdomain and one topdomain followed by 
		an optional port number and an optional path including query information.
	*/
	public static final Pattern URL_REGEXP = Pattern.compile("\\b(http://|https://|ftp://|www.)[\\p{L}\\d_-]+(\\.[\\p{L}\\d_-]+)+(\\:\\d+)?/?[\\p{L}\\d_/.#&+?=;~%-]*");

	/**
		This pattern can be used to find line breaks. It will match any
		combination of carrige return and linefeed characters as well as
		some unicode line separator characters.
	*/
	public static final Pattern LINEBREAKS_REGEXP = Pattern.compile("(\\r\\n)|\\n|\\r|\\u0085|\\u2028|\\u2029");


	/**
		This pattern can be used to fins HTML tags. It will match both
		start and end tags. The entire tag with attributes are put in the
		$1 group, the tag name in the $2 group and the attributes in $3.
	*/
	public static final Pattern TAG_REGEXP = Pattern.compile("<(/?([a-zA-Z]+)\\s*([^>]*))>");


	/**
		This pattern is a list of HTML tags considered "safe".
	*/
	public static final Pattern SAFE_TAGS = Pattern.compile("b|i|ul|ol|li|tt|font|big|small|code", Pattern.CASE_INSENSITIVE);

	/**
		Pattern that matches everything inside a HTML tag.
		@since 2.10
	*/
	public static final Pattern MARKUP = Pattern.compile("<.*?>", Pattern.DOTALL);
	
	/**
		Checks if the given string looks like an email address. This is done
		by trying to match it against the {@link #EMAIL_REGEXP}
		pattern.

		@param email The string to check
		@return TRUE or FALSE
		@see #EMAIL_REGEXP
	*/
	public static boolean isValidEmail(String email)
	{
		return EmailUtil.isValidEmail(email);
	}

	/**
		Checks if the given string looks like an URL. This is done
		by trying to match it against the {@link #URL_REGEXP}
		pattern.

		@param url The string to check
		@return TRUE or FALSE
		@see #URL_REGEXP
	*/
	public static boolean isValidUrl(String url)
	{
		if (url == null) return false;
		Matcher m = URL_REGEXP.matcher(url);
		return m.matches();
	}

	private static final Pattern LEADING_TRAILING_LINEBREAKS = Pattern.compile("^(\\n|\\r|\\u0085|\\u2028|\\u2029)*(.*)(\\n|\\r|\\u0085|\\u2028|\\u2029)*$", Pattern.DOTALL);
	
	/**
		Finds all linebreaks in a string and replaces them with a &lt;br&gt; tag, except
		that leading and trailing linebreaks will be removed.
		
		@param in The string to search
		@return The new string, or an empty string if NULL was passed
	*/
	public static String formatLineBreaks(String in)
	{
		if (in == null) return "";

		// remove leading and trailing CR, LF, CRLF
		Matcher m = LEADING_TRAILING_LINEBREAKS.matcher(in);
		if (m.matches()) in = m.group(2);

		// Replace all CR, LF, CRLF with <br>\n
		m = LINEBREAKS_REGEXP.matcher(in);
		return m.replaceAll("<br>\n");
	}

	/**
		Scans a string for email addresses and URL:s and replaces them
		with <code>&lt;a href="---"&gt;</code> tags.
		Using this method is equivalent to:
		<code>scanForLinks(in, LINK_URL+LINK_EMAIL+SMART_IMAGES, null);</code>
		@param in The string to search
		@return The new string, or an empty string if NULL was passed
	*/
	public static String scanForLinks(String in)
	{
		return scanForLinks(in, LINK_URL+LINK_EMAIL+SMART_IMAGES, null);
	}

	/**
		Scans a string for email addresses and URL:s and replaces them
		with <code>&lt;a href="---"&gt;</code> tags.
		Using this method is equivalent to:
		<code>scanForLinks(in, flags, null);</code>
		@param in The string to search
		@param flags A combination of the following flags to indicate what we should
			search for: {@link #LINK_URL}, {@link #LINK_EMAIL}
			{@link #SMART_IMAGES}
		@return The new string, or an empty string if NULL was passed
	*/
	public static String scanForLinks(String in, int flags)
	{
		return scanForLinks(in, flags, null);
	}

	/**
		Scans a string for email addresses and URL:s and replaces them
		with <code>&lt;a href="---"&gt;</code> tags.
		Using this method is equivalent to:
		<code>scanForLinks(in, LINK_URL+LINK_EMAIL+SMART_IMAGES, target);</code>
		@param in The string to search
		@param target The name of the target window in which the link should be opened
		@return The new string, or an empty string if NULL was passed
	*/
	public static String scanForLinks(String in, String target)
	{
		return scanForLinks(in, LINK_URL+LINK_EMAIL+SMART_IMAGES, target);
	}

	private static final Pattern AT_OR_DOT = Pattern.compile("@|\\.");
	private static final Pattern IMAGE_EXTENSION = Pattern.compile("\\.(jpg|jpeg|gif|png)$");

	/**
		Scans a string for email addresses and URL:s and replaces them
		with <code>&lt;a href="---"&gt;</code> tags.
		@param in The string to search
		@param flags A combination of the following flags to indicate what we should
			search for: {@link #LINK_URL}, {@link #LINK_EMAIL}
			{@link #SMART_IMAGES}
		@param target The name of the target window in which the link should be opened
		@return The new string, or an empty string if NULL was passed
	*/
	public static String scanForLinks(String in, int flags, String target)
	{
		if (in == null) return "";

		if ((flags & LINK_URL) == LINK_URL)
		{
			boolean smartImages = (flags & SMART_IMAGES) == SMART_IMAGES;
			StringBuffer sb = new StringBuffer();
			target = (target == null) ? "" : "\" target=\""+target;
			Matcher m = URL_REGEXP.matcher(in);
			while (m.find())
			{
				String url = m.group(0);
				if (url.startsWith("www")) url = "http://"+url;
				if (smartImages && (IMAGE_EXTENSION.matcher(url).find()))
				{
					m.appendReplacement(sb, "<img src=\""+url+"\" alt=\"*\">");
				}
				else
				{
					m.appendReplacement(sb, "<a href=\""+url+target+"\" rel=\"noopener noreferrer\">$0</a>");
				}
			}
			m.appendTail(sb);
			in = sb.toString();
		}

		if ((flags & LINK_EMAIL) == LINK_EMAIL)
		{
			StringBuffer sb = new StringBuffer();
			Matcher m = EMAIL_REGEXP.matcher(in);
			while (m.find())
			{
				Matcher m2 = AT_OR_DOT.matcher(m.group(0));
				String email = m2.replaceAll("$0<wbr>");
				m.appendReplacement(sb, "<a href=\"mailto:$0\">"+email+"</a>");
			}
			m.appendTail(sb);
			in = sb.toString();
		}
		return in;
	}

	private static final Pattern AMP = Pattern.compile("&");
	private static final Pattern LT = Pattern.compile("<");
	private static final Pattern GT = Pattern.compile(">");
	private static final Pattern QUOTE = Pattern.compile("\"");

	/**
		Scans a string for HTML tags and replaces all &amp; with &amp;amp;,
		&lt; with &amp;lt;, &gt; with &amp;gt; and all " with &#34;
		@param in The string to search
		@return The new string, or an empty string if NULL was passed
	*/
	public static String encodeTags(String in)
	{
		if (in == null) return "";
		in = AMP.matcher(in).replaceAll("&amp;");
		in = LT.matcher(in).replaceAll("&lt;");
		in = GT.matcher(in).replaceAll("&gt;");
		in = QUOTE.matcher(in).replaceAll("&#34;");
		return in;
	}
	
	/**
		Scans a string for HTML tags and replaces all &lt; and &gt; for tags
		not found in the <code>safeTags</code> pattern with &amp;lt; and &amp;gt;
		respectively. Tags that are found in the <code>safeTags</code> pattern
		are not modified. Using this method is equivalent to
		<code>encodeTags(in, Pattern.compile(safeTags));</code>

		@param in The string to search
		@param safeTags A regular expression that should match all
			safe tags
		@return The new string, or an empty string if NULL was passed
		@see #SAFE_TAGS
	*/
	public static String encodeTags(String in, String safeTags)
	{
		if (safeTags == null) return encodeTags(in);
		if (in == null) return "";
		return encodeTags(in, Pattern.compile(safeTags));
	}

	/**
		Scans a string for HTML tags and replaces all &lt; and &gt; for tags
		not found matching the <code>safeTags</code> pattern with &amp;lt; and &amp;gt;
		respectively. Tags that matches the <code>safeTags</code> pattern
		are not modified.

		@param in The string to search
		@param safeTags A regular expression pattern that matches all safe tags
		@return The new string, or an empty string if NULL was passed
		@see #SAFE_TAGS
	*/
	public static String encodeTags(String in, Pattern safeTags)
	{
		if (safeTags == null) return encodeTags(in);
		if (in == null) return "";

		StringBuffer sb = new StringBuffer();
		Matcher m = TAG_REGEXP.matcher(in);
		while (m.find())
		{
			if (safeTags.matcher(m.group(2)).matches())
			{
				if (m.group(1).startsWith("/"))
				{
					m.appendReplacement(sb, "</$2>");
				}
				else
				{
					m.appendReplacement(sb, "<$2>");
				}
			}
			else
			{
				m.appendReplacement(sb, "&lt;$1&gt;");
			}
		}
		m.appendTail(sb);
		return sb.toString();
	}

	private static final Pattern PERCENT = Pattern.compile("\\%");
	private static final Pattern PLUS = Pattern.compile("\\+");
	private static final Pattern SPACE = Pattern.compile("\\s");
	private static final Pattern HASH = Pattern.compile("\\#");
	private static final Pattern AMPERSAND = Pattern.compile("\\&");

	/**
		Encode URL-unsafe characters in a string. Replaces
		% with %25, + with %2B, space with +, # with %23 and &amp; with %26.
		@param in The string to encode
		@return The encoded string, or an empty string if NULL was passed
	*/
	public static String urlEncode(String in)
	{
		if (in == null) return "";
		in = PERCENT.matcher(in).replaceAll("%25");
		in = PLUS.matcher(in).replaceAll("%2B");
		in = SPACE.matcher(in).replaceAll("+");
		in = HASH.matcher(in).replaceAll("%23");
		in = AMPERSAND.matcher(in).replaceAll("%26");
		return in;
	}
	
	private static final Pattern BACKSLASH = Pattern.compile("\\\\");
	private static final Pattern NEWLINE = Pattern.compile("\\r?\\n\\r?");
	private static final Pattern SINGLE_QUOTE = Pattern.compile("\\'");
	private static final Pattern DOUBLE_QUOTE = Pattern.compile("\\\"");
	/**
		Escape a string to make it safe for use in a JavaScript statement.
		Replaces \ with \\, newline with \n, ' with \' and " with \".
	 	@param in String to escape. 
	 	@return a String object, ready to be used in javaScripts. 
	*/
	public static String javaScriptEncode(String in)
	{
		if (in == null) return "";
		in = BACKSLASH.matcher(in).replaceAll("\\\\\\\\");
		in = NEWLINE.matcher(in).replaceAll("\\\\n");
		in = SINGLE_QUOTE.matcher(in).replaceAll("\\\\'");
		in = DOUBLE_QUOTE.matcher(in).replaceAll("\\\\\"");
		return in;
	}

	/**
		A convenience method for doing {@link #encodeTags(String,Pattern) encodeTags(in, SAFE_TAGS)},
		{@link #scanForLinks(String,String) scanForLinks(in, "_blank")} and {@link #formatLineBreaks(String)
		formatLineBreaks(in)} in a single operation.
		@param in The string to format
		@return The result after calling the three methods mentioned above, in that
			order, or an empty string if NULL was passed
	*/
	public static String niceFormat(String in)
	{
		if (in == null) return "";
		return formatLineBreaks(scanForLinks(encodeTags(in, SAFE_TAGS), "_blank"));
	}

	/**
		A convenience method for doing {@link #encodeTags(String,Pattern) encodeTags(in, SAFE_TAGS)},
		{@link #scanForLinks(String,int,String) scanForLinks(in, flags, "_blank")} and
		{@link #formatLineBreaks(String) formatLineBreaks(in)} in a single operation.
		@param in The string to format
		@param flags Flags to be used in the call to {@link #scanForLinks(String,int,String) scanForLinks}
		@return The result after calling the three methods mentioned above, in that
			order, or an empty string if NULL was passed
	*/
	public static String niceFormat(String in, int flags)
	{
		if (in == null) return "";
		return formatLineBreaks(scanForLinks(encodeTags(in, SAFE_TAGS), flags, "_blank"));
	}

	/**
		A convenience method for doing {@link #encodeTags(String,Pattern) encodeTags(in, SAFE_TAGS)},
		{@link #scanForLinks(String,String) scanForLinks(in, target)} and
		{@link #formatLineBreaks(String) formatLineBreaks(in)} in a single operation.
		@param in The string to format
		@param linkTarget The target parameter to be used in the call to
			{@link #scanForLinks(String,String)}
		@return The result after calling the three methods mentioned above, in that
			order, or an empty string if NULL was passed
	*/
	public static String niceFormat(String in, String linkTarget)
	{
		if (in == null) return "";
		return formatLineBreaks(scanForLinks(encodeTags(in, SAFE_TAGS), linkTarget));
	}

	/**
		A convenience method for doing {@link #encodeTags(String,Pattern) encodeTags(in, SAFE_TAGS)},
		{@link #scanForLinks(String,int,String) scanForLinks(in, flags, link_target)} 
		and {@link #formatLineBreaks(String) formatLineBreaks(in)} in a single operation.
		
		@param in The string to format
		@param flags Flags to be used in the call to {@link #scanForLinks(String,int) scanForLinks}
		@param linkTarget The target parameter to be used in the call to
			{@link #scanForLinks(String,int,String)}
		@return The result after calling the three methods mentioned above, in that
			order, or an empty string if NULL was passed
	*/
	public static String niceFormat(String in, int flags, String linkTarget)
	{
		if (in == null) return "";
		return formatLineBreaks(scanForLinks(encodeTags(in, SAFE_TAGS), flags, linkTarget));
	}
	
	/**
		Remove all HTML markup in a string and return what is left.
		@param in The string to strip from HTML
		@return The resulting string, or null if the input is null
		@since 2.10
	*/
	public static String stripMarkup(CharSequence in)
	{
		if (in == null) return null;
		return MARKUP.matcher(in).replaceAll("");
	}
	
	/**
		Counts the length of a string ignoring all characters in HTML markup tags.
		The result is the approximately the length of the string that is displayed 
		on screen by a browser. This method counts all characters that are not
		inside &lt; and &gt;. The actual number of characters displayed by a browser 
		may be less because of escaped sequences, eg. &amp;amp; and white-space that
		is collapsed.
		
		@param html The HTML string
		@return The number of characters that are not HTML markup
		@since 2.10
	*/
	public static int textLength(CharSequence html)
	{
		if (html == null) return 0;
		int length = 0;
		boolean counting = true;
		for (int i = 0; i < html.length(); ++i)
		{
			char c = html.charAt(i);
			if (c == '<') counting = false;
			if (counting) length++;
			if (c == '>') counting = true;
		}
		return length;
	}

}

