/*
	$Id: AbstractHqlQuery.java 7461 2018-03-14 13:17:56Z nicklas $
	
	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.data.BasicData;
import net.sf.basedb.core.query.HqlQuery;
import net.sf.basedb.core.query.QueryParameter;
import net.sf.basedb.core.query.QueryType;
import net.sf.basedb.core.signal.SignalException;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.Map;
import java.util.Collections;
import java.util.Collection;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.hibernate.query.Query;

/**
	An abstract implementation of the {@link HqlQuery} interface. The
	class can be used to build and manage just about any type of HQL
	query. It has no function for executing the query or returning
	the result. Subclasses may use {@link #getMainHqlQuery(DbControl, Class)}
	to generate a Hibernate Query object that is ready to be executed.

	@author Nicklas
	@version 2.15
	@base.modified $Date: 2018-03-14 14:17:56 +0100 (on, 14 mar 2018) $
*/
abstract class AbstractHqlQuery
	extends AbstractQuery
	implements HqlQuery
{

	/**
		Log all parameter bindings to prepared statement.
	*/
	private static final org.slf4j.Logger logParam = 
		org.slf4j.LoggerFactory.getLogger("net.sf.basedb.core.query.bind.AbstractHqlQuery");
		
	/**
		So we don't always have to call logParam.debug()
	*/
	private static final boolean debugEnabled = logParam.isDebugEnabled();

	/**
		The alias for the root table. This is a requirement
		in HQL that is not needed in SQL.
		SELECT ... FROM [rootTable] as [rootAlias]
	*/
	private final String rootAlias;
	
	/**
		If the query should be executed in a stateless session or not.
	*/
	private final boolean stateless;
	
	/**
		The most recent DbControl used to build/execute the main query.
	*/
	private DbControl lastMainDc;

	/**
		The most recent DbControl used to build/execute the count query.
	*/
	private DbControl lastCountDc;
	
	/**
		The most recent DbControl used to build/execute the ID query.
		@since 3.5
	*/
	private DbControl lastIdDc;
	
	/**
		The most recent main query.
	*/
	private org.hibernate.query.Query<?> lastMainQuery;
	
	/**
		The most recent count query.
	*/
	private org.hibernate.query.Query<Long> lastCountQuery;
	
	/**
		The most recent id query.
		@since 3.5
	*/
	private org.hibernate.query.Query<Integer> lastIdQuery;
	
	/**
		If the results should be cached or not.
	*/
	private boolean cacheResults;
	
	/**
		Create a new HQL query. The root table and alias must be given.
	*/
	AbstractHqlQuery(String root, String alias, boolean stateless)
	{
		super(root);
		this.rootAlias = alias;
		this.stateless = stateless;
	}
	
	/*
		From the Query interface
		-------------------------------------------
	*/
	/**
		@return {@link QueryType#HQL}
	*/
	@Override
	public QueryType getQueryType()
	{
		return QueryType.HQL;
	}
	
	/**
		The alias of the item that is the root of this query.
		@see Item#getAlias()
	*/
	@Override
	public String getRootAlias()
	{
		return rootAlias;
	}
	
	@Override
	public boolean isReadonly()
	{
		return super.isReadonly() || lastMainQuery != null;
	}
	
	/**
		Reset all non-permanent query elements of the query and 
		clear cached queries.
	*/
	@Override
	public void reset()
	{
		super.reset();
		lastMainQuery = null;
		lastCountQuery = null;
		lastMainDc = null;
		lastCountDc = null;
	}
	@Override
	public long count(DbControl dc)
		throws BaseException
	{
		long count = -1;
		try
		{
			Query<Long> query = getCountHqlQuery(dc);
			count = isStateless() ?
				QueryExecutor.loadData(Long.class, query, dc.getStatelessSession(), dc.getSessionControl()) :
				HibernateUtil.loadData(query);
		}
		catch (InterruptedException ex)
		{
			throw new SignalException("Aborted by user", ex);
		}
		return count;
	}
	// -------------------------------------------
	/*
		From the HqlQuery interface
		-------------------------------------------
	*/
	@Override
	public void setCacheResult(boolean flag)
	{
		this.cacheResults = flag;
	}
	@Override
	public boolean isCachingResult()
	{
		return cacheResults;
	}
	
	@Override
	public void setEntityParameter(String name, BasicItem value)
	{
		super.setParameter(name, value, null);
	}

	@Override
	public void setPermanentEntityParameter(String name, BasicItem value)
	{
		super.setPermanentParameter(name, value, null);
	}
	// -------------------------------------------
	
	/**
		Build the main query and set parameter values for it. If the query has been 
		executed with the same <code>DbControl</code> before and hasn't been {@link #reset()} 
		the cached query is returned. If the parameters values have been changed the new
		values are used.
		
		@param dc The DbControl to use for executing the query
		@return A <code>org.hibernate.query.Query</code> object
	*/
	@SuppressWarnings("unchecked")
	<R> org.hibernate.query.Query<R> getMainHqlQuery(DbControl dc, Class<R> returnType)
	{
		if (lastMainQuery == null || lastMainDc != dc)
		{
			// If we are using the same DbControl as before we don't have to rebuild the query
			String mainHql = getMainQuery(dc, true);
			if (stateless)
			{
				lastMainQuery =  HibernateUtil.createQuery(dc.getStatelessSession(), mainHql, returnType);
			}
			else
			{
				lastMainQuery =  HibernateUtil.createQuery(dc.getHibernateSession(), mainHql, returnType);
			}
		}
		lastMainDc = dc;
		lastMainQuery.setCacheable(isCachingResult());
		setParameters(lastMainQuery, getParameters());
		if (getFirstResult() >= 0) lastMainQuery.setFirstResult(getFirstResult());
		if (getMaxResults() > 0) lastMainQuery.setMaxResults(getMaxResults());
		return (org.hibernate.query.Query<R>)lastMainQuery;
	}
	
	/**
		Build the count query and set parameter values for it. If the query has been 
		executed with the same <code>DbControl</code> before and hasn't been {@link #reset()} 
		the cached query is returned. If the parameters values have been changed the new
		values are used.
		
		@param dc The DbControl to use for executing the query
		@return A <code>org.hibernate.query.Query</code> object
	*/
	org.hibernate.query.Query<Long> getCountHqlQuery(DbControl dc)
	{
		if (lastCountQuery == null || lastCountDc != dc)
		{
			// If we are using the same DbControl as before we don't have to rebuild the query
			String countHql = getCountQuery(dc, true);
			if (stateless)
			{
				lastCountQuery =  HibernateUtil.createQuery(dc.getStatelessSession(), countHql, Long.class);
			}
			else
			{
				lastCountQuery =  HibernateUtil.createQuery(dc.getHibernateSession(), countHql, Long.class);
			}
		}
		lastCountDc = dc;
		lastCountQuery.setCacheable(true);
		setParameters(lastCountQuery, getParameters());
		return lastCountQuery;
	}
	
	/**
		Build the ID query and set parameter values for it. If the query has been 
		executed with the same <code>DbControl</code> before and hasn't been {@link #reset()} 
		the cached query is returned. If the parameters values have been changed the new
		values are used.
		
		@param dc The DbControl to use for executing the query
		@return A <code>org.hibernate.query.Query</code> object
		@since 3.5
	*/
	org.hibernate.query.Query<Integer> getIdHqlQuery(DbControl dc)
	{
		if (lastIdQuery == null || lastIdDc != dc)
		{
			// If we are using the same DbControl as before we don't have to rebuild the query
			String idHql = getIdQuery(dc, true);
			if (stateless)
			{
				lastIdQuery =  HibernateUtil.createQuery(dc.getStatelessSession(), idHql, Integer.class);
			}
			else
			{
				lastIdQuery =  HibernateUtil.createQuery(dc.getHibernateSession(), idHql, Integer.class);
			}
		}
		lastIdDc = dc;
		lastIdQuery.setCacheable(isCachingResult());
		setParameters(lastIdQuery, getParameters());
		if (getFirstResult() >= 0) lastIdQuery.setFirstResult(getFirstResult());
		if (getMaxResults() > 0) lastIdQuery.setMaxResults(getMaxResults());
		return lastIdQuery;
	}

	/**
		If queries are using the stateless Hibernate session or the regular
		session.
		@return TRUE if the stateless session is used, FALSE otherwise
	*/
	protected boolean isStateless()
	{
		return stateless;
	}

	/**
		Processes the query so that joined paths that are found in other
		parts of the query are replaced with the join alias.
	*/
	@Override
	protected String postProcessQuery(String query) 
	{
		return replaceWithJoinAliases(query);
	}

	/**
		Replaces parts of the query with joined aliases. This is needed
		for Hibernate to generate correct SQL. Eg. 
		select f from Foo f join f.bar b where f.bar.name = :name
		needs to be changed to:
		select f from Foo f join f.bar b where b.name = :name
		<p>
		We implement this by looking for the joined path and replacing it
		with the joined alias.
		<p>
		For left fetch joins we also need to make sure that no more than one
		step is joined at a time since the implicit join created by Hibernate
		is not a fetch join.
	*/
	private String replaceWithJoinAliases(String query)
	{
		// find all 'join foo.bar x' and replace 'foo.bar.' with 'x.'
		Pattern p = Pattern.compile("join(?:\\s+fetch)?\\s+([a-zA-Z0-9\\.]+)\\s+(\\w+)", Pattern.CASE_INSENSITIVE);
		Matcher m = p.matcher(query);
		Set<String> alreadyMatched = new HashSet<String>();
		while (m.find())
		{
			String join = m.group(1);
			String alias = m.group(2);
			if (alreadyMatched.add(alias))
			{
				query = query.replace(join + ".", alias + ".");
				m.reset(query);
				alreadyMatched.add(alias);
			}
		}
		
		// find 'left join fetch' that joins more than one level and replace with individual joins
		// eg. 'left join fetch foo.bar.x' --> 'left join fetch foo.bar as b left join fetch b.x'
		p = Pattern.compile("left\\s+join\\s+fetch\\s+[a-zA-Z0-9]+\\.[a-zA-Z0-9\\.]+(\\.)[a-zA-Z0-9]+\\s+", Pattern.CASE_INSENSITIVE);
		int afj = 1;
		m = p.matcher(query);
		while (m.find())
		{
			int insertJoinAt = m.start(1); // eg. the last dot in 'left join fetch foo.bar.x'
			query = query.substring(0, insertJoinAt) + 
				" AS afj" + afj + " LEFT JOIN FETCH afj" + afj + 
				query.substring(insertJoinAt);
			m = p.matcher(query);
			afj++;
		}
		return query;
	}
	
	
	private static final Set<Integer> ZERO_SET = Collections.singleton(0);

	/**
		Set parameter values on a query.
		@param query The Hibernate query object
		@param parameters A map containing parameter names and values
	*/
	private void setParameters(org.hibernate.query.Query<?> query, Map<String, QueryParameter> parameters)
	{
		if (parameters == null) return;
		if (debugEnabled)
		{
			logParam.debug("Binding parameters to query: " + query.getQueryString());
		}
		for (Map.Entry<String, QueryParameter> entry : parameters.entrySet())
		{
			String name = entry.getKey();
			QueryParameter qp = entry.getValue();
			qp.validateValue();
			Type valueType = qp.getType();
			Item itemType = qp.getItemType();
			Object value = qp.getValue();
			org.hibernate.type.Type hibernateType = null;
			if (itemType != null)
			{
				Class<? extends BasicItem> itemClass = itemType.getItemClass();
				hibernateType = HibernateUtil.getSessionFactory().getTypeHelper().entity(itemType.getDataClass());
				if (itemClass != null)
				{
					// Transform BasicItem:s to BasicData:s
					if (value instanceof Collection)
					{
						Collection<?> c = (Collection<?>)value;
						List<BasicData> tmp = new ArrayList<BasicData>(c.size());
						for (Object o : c)
						{
							tmp.add(itemClass.cast(o).getData());
						}
						value = tmp;
					}
					else
					{
						value = itemClass.cast(value).getData();
					}
				}
			}
			else if (valueType != null)
			{
				hibernateType = valueType.getTypeWrapper().getHibernateType();
			}
			
			if (value instanceof Collection)
			{
				Collection<?> c = (Collection<?>)value;
				if (debugEnabled)
				{
					logParam.debug("Binding collection '"+value+"' to parameter :"+name);
				}
				if (hibernateType == null)
				{
					query.setParameterList(name, c.size() == 0 ? ZERO_SET : c);
				}
				else
				{
					query.setParameterList(name, c.size() == 0 ? ZERO_SET : c, hibernateType);
				}
			}
			else
			{
				if (debugEnabled)
				{
					logParam.debug("Binding '"+value+"' to parameter :"+name);
				}
				if (hibernateType == null)
				{
					query.setParameter(name, value);
				}
				else
				{
					query.setParameter(name, value, hibernateType);
				}
			}
		}
	}

}
