/*
	$Id: AnnotationRestriction.java 7461 2018-03-14 13:17:56Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.core.query.Query;
import net.sf.basedb.core.query.QueryElement;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.util.EqualsHelper;

import org.hibernate.query.NativeQuery;
import org.hibernate.dialect.Dialect;

import java.util.Collection;
import java.util.List;

/**
	Base class for creating restrictions based on annotations.
	Subclasses must implent the {@link #getRestrictionSql(Query, DbControl)}
	and {@link #setRestrictionParameters(NativeQuery)}
	method.
	
	@author Nicklas
	@version 2.0
	@see net.sf.basedb.core.query.Annotations
	@base.modified $Date: 2018-03-14 14:17:56 +0100 (on, 14 mar 2018) $
*/
public abstract class AnnotationRestriction
	implements Restriction
{
	/**
		Log all SQL statements.
	*/
	private static final org.slf4j.Logger logSql = 
		org.slf4j.LoggerFactory.getLogger("net.sf.basedb.core.query.ql");
	
	/**
		So we don't always have to call logSql.debug()
	*/
	private static final boolean debugSqlEnabled = logSql.isDebugEnabled();
	
	private AnnotationType annotationType;
	final Type valueType;
	final int annotationTypeId;
	final Options options;
	final String alias;

	/**
		Create a new annotation restriction for an annotation type.
		@param annotationType The annotation type to use in the query
	*/
	AnnotationRestriction(String alias, AnnotationType annotationType, Options options)
		throws InvalidDataException
	{
		if (annotationType == null) throw new InvalidUseOfNullException("annotationType");
		if (options == null) throw new InvalidUseOfNullException("options");
		if (!options.includePrimary && !options.includeInherited) 
		{
			throw new InvalidDataException("Both includePrimary and includeInherited parameters are FALSE");
		}
		this.annotationType = annotationType;
		this.annotationTypeId = annotationType.getId();
		this.valueType = annotationType.getValueType();
		this.options = options;
		this.alias = alias;
	}

	/**
		Create a new annotation restriction given the id and value type of
		an annotation. It is not verified that there exists an annotation
		type with the specified id or that it's value type is the one specified.
		The only thing that will happen is that the restriction will not match
		any annotations.
		@param annotationTypeId The id of the annotation type
		@param valueType The type of values for annotations
	*/
	@Deprecated
	AnnotationRestriction(String alias, int annotationTypeId, Type valueType, Options options)
		throws InvalidDataException
	{
		if (options == null) throw new InvalidUseOfNullException("options");
		if (!options.includePrimary && !options.includeInherited) 
		{
			throw new InvalidDataException("Both includePrimary and includeInherited parameters are FALSE");
		}
		this.annotationTypeId = annotationTypeId;
		this.valueType = valueType;
		this.options = options;
		this.alias = alias;
	}

	/*
		From the Restriction interface
		-------------------------------------------
	*/
	@SuppressWarnings("unchecked")
	@Override
	public String toQl(Query query, DbControl dc)
		throws BaseException
	{
		SessionControl sc = dc.getSessionControl();
		String valueTable = valueType == null ? null : valueType.getTableName();
		
		// Verify that the annotation type exists
		if (annotationType == null)
		{
			try
			{
				annotationType = AnnotationType.getById(dc, annotationTypeId);
			}
			catch (ItemNotFoundException ex)
			{
				return "1=1";
			}
		}
		int activeProjectId = annotationType.getProjectAnnotations() ? sc.getActiveProjectId() : 0;
		
		Dialect dialect = HibernateUtil.getDialect();
		char oq = dialect.openQuote();
		char cq = dialect.closeQuote();
		
		String valueRestriction = getRestrictionSql(query, dc);
		StringBuilder sb = new StringBuilder();
		/*
			SELECT a.annotationset_id, [a.id], [a.source]
			FROM Annotations a [, valueTable v]
			WHERE a.annotationtype_id = :annotationType
			AND (a.project_id = 0 [OR a.project_id = activeProjectId])
			[AND a.value_id = v.id [AND valueRestriction]]
		*/
		sb.append("SELECT a.[annotationset_id]");
		if (options.includeInherited)
		{
			sb.append(", a.[id], a.[source]");
		}
		sb.append(" FROM [Annotations] a");
		if (valueTable != null)
		{
			sb.append(", [").append(valueTable).append("] v");
		}
		sb.append(" WHERE a.[annotationtype_id] = :annotationType");
		sb.append(" AND (a.[project_id] = ").append(activeProjectId);
		if (activeProjectId != 0)
		{
			if (options.excludeDefaultValues)
			{
				sb.append(" AND a.[override_id]<>0");
			}
			else
			{
				sb.append(" OR (a.[project_id] = 0 AND a.[id] NOT IN (");
				sb.append("select [override_id] from [Annotations] where [project_id]=" + activeProjectId);
				sb.append(" and [annotationtype_id] = " + annotationTypeId);
				sb.append("))");
			}
		}
		sb.append(")");
		if (valueTable != null)
		{
			sb.append(" AND a.[value_id] = v.[id]");
			if (valueRestriction != null)
			{
				sb.append(" AND ").append(valueRestriction);
			}
		}
		if (!options.includeInherited)
		{
			sb.append(" AND a.[source] = ").append(Annotation.Source.PRIMARY.ordinal());
		}
		String sql = sb.toString().replace('[', oq).replace(']', cq);

		org.hibernate.Session session = dc.getHibernateSession();
		NativeQuery<?> sqlQuery = HibernateUtil.createSqlQuery(session, sql);
		sqlQuery.setParameter("annotationType", annotationTypeId, TypeWrapper.H_INTEGER);
		setRestrictionParameters(sqlQuery);
		
		StringBuilder matchingAnnotationSets = new StringBuilder("0");
		if (options.includeInherited)
		{
			if (debugSqlEnabled) logSql.debug("Executing annotation query: " + sql);
			List<Object[]> twoIds = HibernateUtil.loadList((NativeQuery<Object[]>)sqlQuery, sc);
			StringBuilder primarySets = new StringBuilder("0");
			StringBuilder primaryAnnotations = new StringBuilder("0");
			for (Object[] id : twoIds)
			{
				int source = (Integer)id[2];
				if (source == Annotation.Source.PRIMARY.ordinal())
				{
					if (options.includePrimary) matchingAnnotationSets.append(",").append(id[0]);
					primaryAnnotations.append(",").append(id[1]);
				}
				else
				{
					// Cloned annotation
					matchingAnnotationSets.append(",").append(id[0]);
				}
			}

			if (primaryAnnotations.length() > 1)
			{
				// Find inherited annotations to the matching primary annotations
				sql = "SELECT a.[annotationset_id] AS sid"+
					" FROM [Annotations] a"+
					" WHERE a.[inherited_id] IN ("+primaryAnnotations+")"+
					" AND a.[source] = "+Annotation.Source.INHERITED.ordinal();
				
				sql = sql.replace('[', oq).replace(']', cq);
				
				NativeQuery<Integer> sqlQuery2 = HibernateUtil.createSqlQuery(session, sql, Integer.class);
				if (debugSqlEnabled) logSql.debug("Executing annotation query: " + sql);
				List<Integer> ids = HibernateUtil.loadList(sqlQuery2, sc);
				for (Integer id : ids)
				{
					matchingAnnotationSets.append(",").append(id);
				}
			}
			
		}
		else
		{
			if (debugSqlEnabled) logSql.debug("Executing annotation query: " + sql);
			List<Integer> ids = HibernateUtil.loadList((NativeQuery<Integer>)sqlQuery, sc);
			for (Integer id : ids)
			{
				matchingAnnotationSets.append(",").append(id);
			}
		}
		boolean notAmongSelected = notAmongSelected();
		String theAlias = alias != null ? alias : query.getRootAlias();
		StringBuilder hql = new StringBuilder("(");
		hql.append(theAlias).append(".annotationSet");
		if (notAmongSelected) hql.append(" NOT");
		hql.append(" IN (").append(matchingAnnotationSets).append(")");
		if (notAmongSelected)
		{
			hql.append(" OR ").append(theAlias).append(".annotationSet IS NULL");
		}
		hql.append(")");
		return hql.toString();
	}
	/**
		@return Always null
	*/
	@Override
	public Collection<? extends QueryElement> getChildren()
	{
		return null;
	}
	// -----------------------------------------

	/**
		Check as many things as possible to try to determine if the other
		object is equal to this object. This method checks if the other instance
		is the same instance as this object (returns true), if the other instance
		is null or not the same class (returns false), if the other instance has
		a different annotation type, alias or inheritance setting (returns false).
		Subclasses should do more checks if null is returned.
		@return TRUE or FALSE if it is certain or null if more checks are needed
	*/
	protected Boolean maybeEquals(Object other)
	{
		if (this == other) return true;
		if (other == null || this.getClass() != other.getClass()) return false;
		AnnotationRestriction o = (AnnotationRestriction)other;
		if (annotationTypeId != o.annotationTypeId || !EqualsHelper.equals(alias, o.alias) ||
			(options.includeInherited != o.options.includeInherited))
		{
			return false;
		}
		return null;
	}
	@Override
	public int hashCode()
	{
		return EqualsHelper.hashCode(alias) + annotationTypeId;
	}
	// -------------------------------------------

	
	/**
		Get an SQL fragment that includes the actual restriction.
		Use 'v.value' to access the annotation value. Avoid including
		literal values originating from a user, especially strings, since it may lead 
		to security holes unless the values are checked
		for dangerous characters. Use a parameter placeholder instead. A parameter
		placeholder is a colon (:) followed by a name.
		Example:
		<pre class="code">
v.value = :theValue
</pre>
		The <code>setRestrictionParameters</code> method is called to 
		allow the subclass to set the value of the parameters specified
		in the SQL fragment.

		@param query The query object we are about to execute
		@return An SQL fragment, or null if no additional restriction is required
		@see #setRestrictionParameters(NativeQuery)
	*/
	abstract String getRestrictionSql(Query query, DbControl dc)
		throws BaseException;
	
	/**
		Set the value for all parameters defined in the SQL fragment
		returned by {@link #getRestrictionSql(Query, DbControl)}.
		@param query The Hibernate <code>SQLQuery</code> object which is
			used to query the database
	*/
	abstract void setRestrictionParameters(NativeQuery<?> query);
	
	/**
		If the main restriction should match items that had annotations 
		among the selected annotation sets, or not.
		@return FALSE (default) to match the selected annotation sets, TRUE
			to match the non-selected annotation sets
		@since 2.7
	*/
	boolean notAmongSelected()
	{
		return false;
	}
	
	/**
		Options that specify which annotations to consider when searching.
		@since 3.11
	*/
	public static class Options
	{
		private boolean includePrimary;
		private boolean includeInherited;
		private boolean excludeDefaultValues;
		
		/**
			Creates the default options.
		*/
		public Options()
		{
			this.includePrimary = true;
			this.includeInherited = false;
			this.excludeDefaultValues = false;
		}
		
		/**
			Should primary annotations be included in the search or not.
			Enabled by default.
		*/
		public boolean includePrimary()
		{
			return includePrimary;
		}
		
		public Options setIncludePrimary(boolean includePrimary)
		{
			this.includePrimary = includePrimary;
			return this;
		}
	
		/**
			Should inherited annotations be included in the search or not.
			Disabled by default.
		*/
		public boolean includeInherited()
		{
			return includePrimary;
		}
		
		public Options setIncludeInherited(boolean includeInherited)
		{
			this.includeInherited = includeInherited;
			return this;
		}

		/**
			Should default values (for project-specific annotation types)
			be excluded from the query or not. Disabled by default.
		*/
		public boolean excludeDefaultValues()
		{
			return excludeDefaultValues;
		}
		
		public Options setExcludeDefaultValues(boolean excludeDefaultValues)
		{
			this.excludeDefaultValues = excludeDefaultValues;
			return this;
		}
	
	}
	
}

