/**
	$Id: LoggingInterceptor.java 7132 2016-04-21 11:13:30Z nicklas $

	Copyright (C) 2009 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.log;

import java.io.IOException;
import java.io.NotSerializableException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;



import net.sf.basedb.core.LogControl;
import net.sf.basedb.util.extensions.ExtensionsInvoker;

import org.hibernate.EmptyInterceptor;
import org.hibernate.SessionFactory;
import org.hibernate.Transaction;
import org.hibernate.resource.transaction.spi.TransactionStatus;
import org.hibernate.type.Type;

/**
	Interceptor that attaches to Hibernate sessions to provide
	logging of items that are updated/created/deleted. Do not attach
	this interceptor to a {@link SessionFactory} since this class is
	not thread-safe.
	<p>
	The interceptor uses <code>onSave</code>, <code>onDelete</code> and
	<code>onFlushDirty</code> to collect information about what has happened.
	The {@link LogManagerFactory} is asked to create {@link EntityLogger}:s
	for each modified/created/deleted item. The actual logging takes place
	in the <code>postFlush</code> method that calls {@link 
	EntityLogger#logChanges(LogManager, EntityDetails)} on each entity logger.
	<p>
	NOTE! For new items, the ID is not available until after the flush
	to the database has been done. Eg. new items has a 0 id when 
	{@link LogManagerFactory#getEntityLogger(LogManager, Object)} is called.
	
	@author Nicklas
	@version 2.13
	@base.modified $Date: 2016-04-21 13:13:30 +0200 (to, 21 apr 2016) $
*/
public class LoggingInterceptor
	extends EmptyInterceptor
{

	private static final long serialVersionUID = -9108587050205509572L;
	
	// Empty arrays to use when there is no other information
	private static final Object[] EMTPY_OBJECT_ARRAY = new Object[0];
	private static final String[] EMTPY_STRING_ARRAY = new String[0];
	private static final Type[] EMTPY_TYPE_ARRAY = new Type[0];
	
	private final LogControl logControl;
	private final ExtensionsInvoker<LogManagerFactory> logInvoker;
	private List<LogManagerAndFactory> logHandlers;
	private List<LogDetails> changes;
	
	/**
		Creates a new interceptor.
		@param logControl The LogControl that is managing the current transaction
		@param logInvoker The extension invoker that create log manager factories
	*/
	public LoggingInterceptor(LogControl logControl, ExtensionsInvoker<LogManagerFactory> logInvoker)
	{
		this.logControl = logControl;
		this.logInvoker = logInvoker;
	}
	
	/*
		From the Interceptor interface
		--------------------------------
	*/
	/**
		Adds a CREATE event to the log.
		@return Always false, since the state is not modified
	*/
	@Override
	public boolean onSave(Object entity, Serializable id, Object[] state,
			String[] propertyNames, Type[] types)
	{
		addChange(ChangeType.CREATE, entity, state, null, propertyNames, types, false);
		return false;
	}
	
	/**
		Adds an UPDATE event to the log.
		@return Always false, since the state is not changed
	*/
	@Override
	public boolean onFlushDirty(Object entity, Serializable id,
			Object[] currentState, Object[] previousState, String[] propertyNames,
			Type[] types)
	{
		addChange(ChangeType.UPDATE, entity, currentState, previousState, propertyNames, types, false);
		return false;
	}
	
	/**
		Adds a DELETE event to the log.
	*/
	@Override
	public void onDelete(Object entity, Serializable id, Object[] state,
			String[] propertyNames, Type[] types)
	{
		addChange(ChangeType.DELETE, entity, state, state, propertyNames, types, false);
	}

	/**
		Collects all log events and sends them to the each {@link EntityLogger}.
	*/
	@SuppressWarnings("rawtypes")
	@Override
	public void postFlush(Iterator entities)
	{
		if (changes != null && changes.size() > 0)
		{
			for (LogDetails entry : changes)
			{
				entry.logger.logChanges(entry.manager, entry.details);
			}
			// NOTE! A flush may happen more than once during the lifetime of a
			// transaction, so we must clear the list of changes
			changes.clear();
		}
	}

	/**
		Flush remaining changes in the log control.
	 */
	@Override
	public void beforeTransactionCompletion(Transaction tx) 
	{
		logControl.flush();
	}

	/**
		Signals successful commit or rollback to the log manager.
		Note that this happens after the transaction/connection has been
		closed and no communication with the database can take place
		here.
	*/
	@Override
	public void afterTransactionCompletion(Transaction tx)
	{
		if (logHandlers != null)
		{
			boolean rollback = tx.getStatus() != TransactionStatus.COMMITTED;
			for (LogManagerAndFactory handler : logHandlers)
			{
				try
				{
					if (rollback)
					{
						handler.manager.afterRollback();
					}
					else
					{
						handler.manager.afterCommit();
					}
				}
				catch (RuntimeException ex)
				{}
			}
		}
		logHandlers = null;
		changes = null;
	}
	// --------------------------------------------

	/**
		Adds a manual non-Hibernate change to the log system. The change is
		forwarded to the log handlers immediately. The
		log handlers should not store the entity object for later use since
		instance may be reused for other changes in the future.
		@since 3.3
	*/
	public void addManualSingleChange(ChangeType action, Object entity, Object newValue, Object oldValue, String property, Type type)
	{
		addChange(action, entity, new Object[] { newValue }, new Object[] { oldValue }, new String[] { property }, new Type[] { type }, true);
	}
	
	/**
		Adds a manual non-Hibernate change to the log system. The entity should be
		a special object that doesn't rely on the state arrays to convey what has
		changed. The change is forwarded to the log handlers immediately. The
		log handlers should not store the entity object for later use since
		instance may be reused for other changes in the future.
		@since 3.8
	*/
	public void addManualSingleChange(ChangeType action, Object entity)
	{
		addChange(action, entity, EMTPY_OBJECT_ARRAY, EMTPY_OBJECT_ARRAY, EMTPY_STRING_ARRAY, EMTPY_TYPE_ARRAY, true);
	}
	
	/**
		If the log manager factory creates an entity logger, add
		information about the change to the internal list. The list
		is processed in the {@link #postFlush(Iterator)} method.
	 */
	private void addChange(ChangeType action, Object entity, Object[] state, 
			Object[] previousState, String[] propertyNames, Type[] types, boolean logImmediately)
	{
		if (logHandlers == null)
		{
			// First detected change, create log managers and space for recording changes
			this.logHandlers = new ArrayList<LogManagerAndFactory>(logInvoker.getNumExtensions());
			this.changes = new LinkedList<LogDetails>();
			
			for (LogManagerFactory factory : logInvoker)
			{
				logHandlers.add(new LogManagerAndFactory(factory, factory.getLogManager(logControl)));
			}
			
		}
		
		EntityDetails details = null;
		for (LogManagerAndFactory handler : logHandlers)
		{
			EntityLogger logger = handler.factory.getEntityLogger(handler.manager, entity);
			if (logger != null)
			{
				if (details == null) details = new EntityDetails(entity, action, 
						state, previousState, propertyNames, types);
				if (logImmediately)
				{
					logger.logChanges(handler.manager, details);	
				}
				else
				{
					changes.add(new LogDetails(handler.manager, logger, details));
				}
			}
		}
	}
	
	// This class is not serializable
	private void readObject(ObjectInputStream ois)
		throws ClassNotFoundException, IOException 
	{
		throw new NotSerializableException();
	}
	private void writeObject(ObjectOutputStream ois)
		throws IOException 
	{
		throw new NotSerializableException();
	}

	
	static class LogManagerAndFactory
	{
		final LogManagerFactory factory;
		final LogManager manager;
		
		LogManagerAndFactory(LogManagerFactory factory, LogManager manager)
		{
			this.factory = factory;
			this.manager = manager;
		}
	}
	
	static class LogDetails
	{
		final LogManager manager;
		final EntityLogger logger;
		final EntityDetails details;
		
		LogDetails(LogManager manager, EntityLogger logger, EntityDetails details)
		{
			this.manager = manager;
			this.logger = logger;
			this.details = details;
		}
	}
}
