/**
	$Id $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.plugin;

import java.io.Serializable;
import java.util.List;

import net.sf.basedb.core.ParameterException;

/**
	A helper class that wraps a request object with a namespace. 
	All parameter reading operations will be made to the specified
	namespace. Eg, the 'name' parameter is replaced with the namespace 
	+ dot + the original name. For example, if namespace="export" and
	name="toDir", the resulting name is "export.toDir". If the parameter
	doesn't exists, the global namespace is checked. (eg. toDir).
	All other methods simply forward to the same method of the parent 
	object.
	<p>
	The main usage for this class is when a master plug-in needs to use
	other child plug-ins. By wrapping the request with an instance 
	of this class it makes it possible to forward the request to the child
	plug-in without having to worry about name clashes among the children.
	
	@author nicklas
	@since 2.15
	@base.modified $Date $
	@see NamespaceParameterValuesWrapper
	@see NamespacePluginParameter
*/
public class NamespaceRequestWrapper
	implements Request
{

	private final Request parent;
	private final String namespace;
	
	/**
		Create a new namespace wrapper for the parent request.
		@param parent The parent request values, null is not allowed
		@param namespace The namespace, null is not allowed
	*/
	public NamespaceRequestWrapper(Request parent, String namespace)
	{
		if (parent == null) throw new NullPointerException("parent");
		if (namespace == null) throw new NullPointerException("namespace");
		this.parent = parent;
		this.namespace = namespace;
	}
	
	@Override
	public String getCommand()
	{
		return parent.getCommand();
	}
	
	@Override
	public Serializable getState()
	{
		return parent.getState();
	}

	@Override
	public Object getParameterValue(String name) 
		throws ParameterException
	{
		Object value = parent.getParameterValue(namespace + "." + name);
		if (value == null) value = parent.getParameterValue(name);
		return value;
	}

	@Override
	public List<?> getParameterValues(String name)
	{
		List<?> values = parent.getParameterValues(namespace + "." + name);
		if (values == null) values = parent.getParameterValues(name);
		return values;
	}

	@Override
	public boolean isAllowedImmediateExecution()
	{
		return parent.isAllowedImmediateExecution();
	}

}
