/*
	$Id: Expressions.java 7034 2015-12-04 07:47:16Z nicklas $

	Copyright (C) 2005 Samuel Andersson, Nicklas Nordborg
	Copyright (C) 2006 Johan Enell, Jari Häkkinen, Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.query;

import net.sf.basedb.core.Type;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.NumberOutOfRangeException;
import net.sf.basedb.util.MD5;
import net.sf.basedb.util.filter.Filter;

import java.util.regex.Pattern;

/**
	A factory class to create expressions.

	@author Samuel, Nicklas
	@version 2.0
	@base.modified $Date: 2015-12-04 08:47:16 +0100 (fr, 04 dec 2015) $
*/
public class Expressions
{

	/**
		A parameter name can only contain the characters a-z, A-Z or 0-9.
	*/
	public static final Pattern PARAMETER_REGEXP = Pattern.compile("[a-zA-z0-9]+");

	/**
		Add two expressions: new expression = e1 + e2.
		@param e1 The left value
		@param e2 The right value
		@return The new expression
		@throws InvalidDataException If any of the arguments are null
	*/
	public static Expression add(Expression e1, Expression e2)
		throws InvalidDataException
	{
		if (e1 == null) throw new InvalidUseOfNullException("e1");
		if (e2 == null) throw new InvalidUseOfNullException("e2");
		return new AddExpression(e1, e2);
	}

	/**
		Subtract one expression from another: new expression = e1 - e2.
		@param e1 The left value
		@param e2 The right value
		@return The new expression
		@throws InvalidDataException If any of the arguments are null
	*/
	public static Expression subtract(Expression e1, Expression e2)
		throws InvalidDataException
	{
		if (e1 == null) throw new InvalidUseOfNullException("e1");
		if (e2 == null) throw new InvalidUseOfNullException("e2");
		return new SubtractExpression(e1, e2);
	}

	/**
		Divide one expression by another: new expression = e1 / e2.
		@param e1 The numerator
		@param e2 The denominator
		@return The new expression
		@throws InvalidDataException If any of the arguments are null
	*/
	public static Expression divide(Expression e1, Expression e2)
		throws InvalidDataException
	{
		if (e1 == null) throw new InvalidUseOfNullException("e1");
		if (e2 == null) throw new InvalidUseOfNullException("e2");
		return new DivideExpression(e1, e2);
	}

	/**
		Multiply two expressions: new expression = e1 * e2.
		@param e1 The left value
		@param e2 The right value
		@return The new expression
		@throws InvalidDataException If any of the arguments are null
	*/
	public static Expression multiply(Expression e1, Expression e2)
		throws InvalidDataException
	{
		if (e1 == null) throw new InvalidUseOfNullException("e1");
		if (e2 == null) throw new InvalidUseOfNullException("e2");
		return new MultiplyExpression(e1, e2);
	}

	/**
		Negate an expression: new expression = -e1.
		@param e The expression to negate
		@return The new expression
		@throws InvalidDataException If the argument is null
	*/
	public static Expression negate(Expression e)
		throws InvalidDataException
	{
		if (e == null) throw new InvalidUseOfNullException("e");
		return new NegateExpression(e);
	}

	/**
		Create a constant expression from an integer.
		@param value The value to create an expression for
		@return The expression
	*/
	public static Expression integer(int value)
	{
		return new IntegerExpression(value);
	}

	/**
		Create a constant expression from a float.
		@param value The value to create an expression for
		@return The expression
	*/
	public static Expression aFloat(float value)
	{
		return new FloatExpression(value);
	}

	/**
		Create a constant expression from a string. Internally this is implemented
		as parameter with a default value. Ie. {@link #parameter(String, Object)}.
		In most cases it is better to use that method instead of this.
		@param value The string value
		@return The expression
		@throws InvalidDataException If the value is null
	*/
	public static Expression string(String value)
		throws InvalidDataException
	{
		if (value == null) throw new InvalidUseOfNullException("value");
		return parameter("P" + MD5.getHashString(value), value);
	}
	
	/**
		Create a constant expression from a boolean.
		@param value The value to create an expression for
		@return The expression
		@since 3.1
	*/
	public static Expression bool(boolean value)
	{
		return new BooleanExpression(value);
	}
	
	/**
		Create a boolean expression view of a restriction.
		@param restriction The restriction
		@return The expression
		@since 3.1
	*/
	public static Expression bool(Restriction restriction)
	{
		if (restriction == null) throw new InvalidUseOfNullException("restriction");
		return new BooleanExpression(restriction);
	}
	
	/**
		Create an expression for a parameter without a default value or type.
	 	@param name Name of the parameter. Can not be null.
	 	@return An expression.
	 	@throws InvalidDataException If the parameter is null. 
		@see #parameter(String, Object, Type)
		@see Query#setParameter(String, Object, Type)
	*/
	public static Expression parameter(String name)
		throws InvalidDataException
	{
		return parameter(name, null, null);
	}
	
	/**
		Create an expression for a typed parameter without a default value.
	 	@param name Name of the parameter. Can not be null 
	 	@param valueType Type of value the parameter holds.
	 	@return An expression
	 	@throws InvalidDataException If required argument is null. 
		@see #parameter(String, Object, Type)
		@see Query#setParameter(String, Object, Type)
	*/
	public static Expression parameter(String name, Type valueType)
		throws InvalidDataException
	{
		return parameter(name, null, valueType);
	}

	/**
		Create an expression for a parameter with a default value but no type.
	 	@param name Name of the parameter. Can not be null.
	 	@param value Value of the parameter.
	 	@return An expression to use in queries.
	 	@throws InvalidDataException If required parameter is null.
		@see #parameter(String, Object, Type)
	*/
	public static Expression parameter(String name, Object value)
		throws InvalidDataException
	{
		return parameter(name, value, null);
	}

	/**
		Create an expression for a parameter with a default value and type.
		@param name The name of the parameter. Can not be null
		@param value The default value of the parameter, or null if no default is needed
		@param valueType The type of the parameter, or null if not used
		@return An expression representing the parameter
		@throws InvalidDataException If the argument 'name' is null
			or contains invalid characters
		@see Query#setParameter(String, Object, Type)
		@see #PARAMETER_REGEXP
	*/
	public static Expression parameter(String name, Object value, Type valueType)
		throws InvalidDataException
	{
		if (name == null) throw new InvalidUseOfNullException("name");
		if (!PARAMETER_REGEXP.matcher(name).matches()) 
		{
			throw new InvalidDataException("Parameter '"+name+"' has one or more invalid characters. Only a-z, A-Z and 0-9 are allowed.");
		}
		return new ParameterExpression(name, value, valueType);
	}
	
	/**
		Create an expression for an already selected query element. If an
		alias has been specified it will be used, otherwise the entire expression
		is built again.
		@param select A selected query element. Can not be null.
		@return A expression object representing the selected expression
		@throws InvalidDataException If the argument is null
	*/
	public static Expression selected(Select select)
		throws InvalidDataException
	{
		if (select == null) throw new InvalidUseOfNullException("select");
		return new SelectedExpression(select);
	}
	
	/**
		Calculate the 10-based logarithm of an expression: new expression = log10(e).
		@param e The expression to take the logarithm of
		@return The new expression
		@throws InvalidDataException If the argument is null
	*/
	public static Expression log10(Expression e)
		throws InvalidDataException
	{
		if (e == null) throw new InvalidUseOfNullException("e");
		return new LogExpression(10.0, e);
	}
	
	/**
		Calculate the 2-based logarithm of an expression: new expression = log2(e).
		@param e The expression to take the logarithm of
		@return The new expression
		@throws InvalidDataException If the argument is null
	*/
	public static Expression log2(Expression e)
		throws InvalidDataException
	{
		if (e == null) throw new InvalidUseOfNullException("e");
		return new LogExpression(2.0, e);
	}

	/**
		Calculate the natural logarithm of an expression: new expression = ln(e).
		@param e The expression to take the logarithm of
		@return The new expression
		@throws InvalidDataException If the argument is null
	*/
	public static Expression ln(Expression e)
		throws InvalidDataException
	{
		if (e == null) throw new InvalidUseOfNullException("e");
		return new LogExpression(Math.E, e);
	}
	
	/**
		Calculate the n-based logarithm of an expression: new expression = log(n, e).
		@param n The log base, must be &gt; 0
		@param e The expression to take the logarithm of
		@return The new expression
		@throws InvalidDataException If the expression is null or if the base is 0 or
			below
	*/
	public static Expression log(double n, Expression e)
		throws InvalidDataException
	{
		if (n <= 0.0) throw new NumberOutOfRangeException("n", n, 0.0, false);
		if (e == null) throw new InvalidUseOfNullException("e");
		return new LogExpression(n, e);
	}

	/**
		Calculate the exponential of an expression: new expression = exp(e).
		@param e The expression to take the exponential of
		@return The new expression
		@throws InvalidDataException If the argument is null
	*/
	public static Expression exp(Expression e)
		throws InvalidDataException
	{
		if (e == null) throw new InvalidUseOfNullException("e");
		return new ExpExpression(e);
	}
	
	/**
		Calculate the first expression raised to the power
		of the second expression: new expression = e1 ^ e2
		@param e1 The base expression
		@param e2 The exponent expression
		@return The new expression
		@since 2.12
		@throws InvalidDataException If some of the arguments are null
	*/
	public static Expression power(Expression e1, Expression e2)
	{
		if (e1 == null) throw new InvalidUseOfNullException("e1");
		if (e2 == null) throw new InvalidUseOfNullException("e2");
		return new PowerExpression(e1, e2);		
	}
	
	/**
		Calculate the square root of an expression: new expression = sqrt(e).
		@param e The expression to take the square root of
		@return The new expression
		@throws InvalidDataException If the argument is null
	*/
	public static Expression sqrt(Expression e)
		throws InvalidDataException
	{
		if (e == null) throw new InvalidUseOfNullException("e");
		return new SqrtExpression(e);
	}
	
	/**
		Calculate the absolute value of an expression: new expression = abs(e).
		@param e The expression to take the absolute value of
		@return The new expression
		@throws InvalidDataException If the argument is null
	*/
	public static Expression abs(Expression e)
		throws InvalidDataException
	{
		if (e == null) throw new InvalidUseOfNullException("e");
		return new AbsExpression(e);
	}

	/**
		Get the date part from a date or timestamp expression.
		@param e An expression which should evaluate to a date or timestamp
		@return The new expression
		@since 2.16
	*/
	public static Expression toDate(Expression e)
	{
		if (e == null) throw new InvalidUseOfNullException("e");
		return new ToDateExpression(e);
	}
	
	/**
		Create a conditional <code>CASE-WHEN</code> expression:
		<pre class="code">
CASE
WHEN r1 THEN e1
WHEN r2 THEN e2
...
ELSE elseExpression
END
</pre>
		@param elseExpression The optional else part of the expression
		@param whenStatements An array of <code>WhenStatement</code>:s
		@return The new expression
		@throws InvalidDataException If any of the when statements are null or
			if no when statements have been specified
	*/
	public static Expression caseWhen(Expression elseExpression, WhenStatement... whenStatements)
		throws InvalidDataException
	{
		int firstNull = getFirstNull(whenStatements);
		if (firstNull >= 0 || whenStatements == null)
		{
			throw new InvalidUseOfNullException("whenStatements["+firstNull+"]");
		}
		if (whenStatements.length < 1) 
		{
			throw new InvalidDataException("'whenStatements' must contain at least one element");
		}
		return new CaseExpression(elseExpression, whenStatements);
	}
	
	/**
		Create a subquery expression: <code>ANY (subquery)</code>.
		The expression is useful in the WHERE part of a query only.
		For example, to find all users that are members of a group
		having a name starting with A.
		<pre class="code">
Query groupFilter = User.getQuery();
groupFilter.join(Hql.innerJoin("groups", "grp"));
groupFilter.restrict(Restrictions.like(Hql.property("grp", "name"), Expressions.string("A%")));
Query userQuery = User.getQuery();
userQuery.restrict(Restrictions.eq(Hql.property("id"), Expressions.any(groupFilter)));
</pre>
		
		@param subquery The subquery
		@return The new expression
		@throws InvalidDataException If the subquery is null
		@since 2.9
	*/
	public static Expression any(Query subquery)
	{
		if (subquery == null) throw new InvalidUseOfNullException("subquery");
		return new SubqueryExpression("ANY", subquery);
	}
	
	/**
		Create a subquery expression: <code>ANY (subquery)</code>.
		WARNING!!! It is not recommended to use this method for creating
		restriction like <code>&lt;&gt;ANY (subquery)</code> since it may
		have a bad impact on performance. Consider using 
		{@link Restrictions#not(Restriction)} and
		{@link Restrictions#in(Expression, Query)} instead.
		
		@param subquery The subquery
		@return The new expression
		@throws InvalidDataException If the subquery is null
		@see #any(Query)
		@since 2.9
	*/
	public static Expression all(Query subquery)
	{
		if (subquery == null) throw new InvalidUseOfNullException("subquery");
		return new SubqueryExpression("ALL", subquery);
	}
	
	/**
		Create an expression that falls back to one of two expression
		depending on if the query passes a filter or not.
		
		@param filter A filter, if null the 'ifTrue' expression is returned
		@param ifTrue The epxression to use if the filter evaluates to TRUE
		@param ifFalse The expression to use if the filter evaluates to FALSE
		@return A conditional expression, or the 'ifTrue' expression if the filter is null, or
			null if both expressions are null
		@since 3.7
	*/
	@SuppressWarnings("unchecked")
	public static Expression conditionalExpression(Filter<? extends Query> filter, Expression ifTrue, Expression ifFalse)
	{
		if (filter == null || (ifTrue == null && ifFalse == null)) return ifTrue;
		return new ConditionalQueryElement((Filter<Query>)filter, ifTrue, ifFalse);
	}
	
	/**
		Test if there is null in the array.
		@return The index of the first null element,
			or -1 if no null elements are found, or 0 if
			the array itself is null
	*/
	private static int getFirstNull(Object[] array)
	{
		if (array == null) return 0;
		for (int i = 0; i < array.length; ++i)
		{
			if (array[i] == null) return i;
		}
		return -1;
	}

}
