/*
	$Id: Hql.java 7105 2016-03-10 08:50:45Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Johan Enell

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.query;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ReporterList;
import net.sf.basedb.core.Shareable;
import net.sf.basedb.core.UnsavedItemException;
import net.sf.basedb.core.data.BasicData;

import java.util.regex.Pattern;

/**
	A factory class to create expressions, joins, etc. that are only used
	by HQL queries. The elements will throw an
	<code>UnsupportedOperationException</code> if passed to
	a SQL query.

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2016-03-10 09:50:45 +0100 (to, 10 mar 2016) $
*/
public class Hql
{

	/**
		Log core events.
	*/
	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(Hql.class);
	
	
	/**
		A property can only contain the characters a-zA-Z0-9, period(.), underscore(_) or hash (#)
		It cannot begin or end with a period/hash and cannot have two periods/hashes in a sequence.
		A hash indicates a path to a component while a period indicates a path to another
		entity.
	*/
	public static final Pattern PROPERTY_REGEXP = Pattern.compile("[a-zA-Z0-9_]+((\\.|\\#)[a-zA-Z0-9_]+)*");

	/**
		An alias can only contain the characters a-z, A-Z or 0-9.
	*/
	public static final Pattern ALIAS_REGEXP = Pattern.compile("[a-zA-z0-9]+");

	/**
		Create an expression representing an item which is a subclass of BasicItem. 
		In the query the actual value used is the id of the entity. 
		See {@link #property(String, String)} for a code example.

		@param entity The entity
	 	@return an object implementing {@link net.sf.basedb.core.query.Expression} 
		@throws InvalidDataException If the entity is null or isn't saved to the database
		@see #entity(BasicData)
	*/
	public static Expression entity(BasicItem entity)
		throws InvalidDataException
	{
		if (entity == null) throw new InvalidUseOfNullException("entity");
		if (!entity.isInDatabase()) 
		{
			throw new UnsavedItemException("Can't use an unsaved item in query: " + entity);
		}
		return new HqlEntityExpression(entity);
	}
	
	/**
		Create an expression representing an item which is a subclass of BasicData,
		for example a reporter. In the query the actual value used is the id of the entity. 
		See {@link #property(String, String)} for a code example.
	
		@param entity The entity
		@throws InvalidDataException If the entity is null
		@see #entity(BasicItem)
		@since 2.2.2
	*/
	public static Expression entity(BasicData entity)
		throws InvalidDataException
	{
		if (entity == null) throw new InvalidUseOfNullException("entity");
		if (entity.getId() == 0) 
		{
			throw new UnsavedItemException("Can't use an unsaved item in query: " + entity);
		}
		return new HqlEntityExpression(entity);
	}
	
	/**
		Create an expression for a parameter which must be an entity
		of a specified type.
		@param name The name of the parameter. Can not be null
		@param itemType The type of the
		@return An expression representing the parameter
		@throws InvalidDataException If the 'name' is null
			or contains invalid characters or if 'itemType' is null
		@since 2.17
	*/
	public static Expression entityParameter(String name, Item itemType)
	{
		if (name == null) throw new InvalidUseOfNullException("name");
		if (itemType == null) throw new InvalidUseOfNullException("itemType");
		return new HqlEntityParameterExpression(name, itemType);
	}
	
	/**
		Same as <code>property(null, property)</code>.
		@see #property(String, String)
	*/
	public static Expression property(String property)
		throws InvalidDataException
	{
		return property(null, property, null);
	}
	
	/**
		Same as <code>property(alias, null)</code>.
		@see #property(String, String)
	*/
	public static Expression alias(String alias)
		throws InvalidDataException
	{
		return property(alias, null, null);
	}

	/**
		Create an expression representing a property of an object or joined alias.
		Property names are usually named after a getter method minus the
		"get" part. Ie. a <code>getName()</code> method corresponds to
		a "name" property. If no alias is given the property name is resolved 
		against the root entity of the query, otherwise it is resolved against the
		joined entity with the given alias.
<pre class="code">
<b>// Sorting user by the "name" property</b>
ItemQuery&lt;User&gt; query = User.getQuery();
query.order(Orders.asc(Hql.property("name")));

<b>// Generated HQL</b>
SELECT usr
FROM UserData usr
ORDER BY usr.name ASC

<b>// Loading users members of the Administrators role</b>
DbControl dc = ...
int adminId = SystemItems.get(Role.ADMINISTRATOR); // Assume adminId = 1
Role admin = Role.getById(dc, adminId);
ItemQuery&lt;User&gt; query = User.getQuery();
query.join(Hql.innerJoin("roles", "rle"));
query.restrict(
  Restrictions.eq(
    Hql.alias("rle"),
    Hql.entity(admin)
  )
);

<b>// Generated HQL</b>
SELECT usr 
FROM UserData usr
JOIN usr.roles rle
WHERE rle = 1
</pre>
		
		@param property The property name of the object
		@param alias Alias to resolve the property against, use
			null to resolve the property agains the root entity
		@return An expression representing the property
		@throws InvalidDataException If both the property and alias is null or 
			any of the parameters contains invalid characters
	*/
	public static Expression property(String alias, String property)
		throws InvalidDataException
	{
		return property(alias, property, null);
	}

	/**
		Create an expression representing the value of an indexed property.
		An indexed property is an element in a list, array or map. The alias
		and property is used as in {@link #property(String, String)}. The index
		expression is applied as: <code>property[index]</code>.
		
		@param property The property name of the object
		@param alias Alias to resolve the property against, use
			null to resolve the property agains the root entity
		@param index An index expression (null is allowed)
		@return An expression representing the indexed property
		@since 3.8
	*/
	public static Expression property(String alias, String property, Expression index)
	{
		if (property == null && alias == null) throw new InvalidUseOfNullException("property and alias");
		if (property != null)
		{
			if (!PROPERTY_REGEXP.matcher(property).matches()) 
			{
				throw new InvalidDataException("Property '"+property+"' has one or more invalid characters. Only a-z, A-Z, 0-9, _ and . is allowed.");
			}
			if (property.endsWith("affyChip"))
			{
				log.warn("Property 'affyChip' has been deprecated. " +
						"Use 'platform.externalId' and compare it to Platform.AFFYMETRIX");
			}
		}
		if (alias != null && !ALIAS_REGEXP.matcher(alias).matches()) 
		{
			throw new InvalidDataException("Alias '"+alias+"' has one or more invalid characters. Only a-z, A-Z and 0-9 is allowed.");
		}
		return new HqlPropertyExpression(alias, property, index);
	}
	
	/**
		Create an index expression for a property that is a map or
		a list. If no alias is given the property name is resolved 
		against the root entity of the query, otherwise it is resolved against the
		joined entity with the given alias. If no property is given only the alias
		is used.

		@param alias The alias to resolve the property agains, or null to resolve the
			property against the root entity
		@param property The property name of the object which must be a map or list association
		@return An expression representing the index collection of the property
		@throws InvalidDataException If both the alias and property parameters are null
			or if any of them contains invalid characters
	*/
	public static Expression index(String alias, String property)
		throws InvalidDataException
	{
		if (alias == null && property == null) throw new InvalidUseOfNullException("alias & property");
		if (property != null && !PROPERTY_REGEXP.matcher(property).matches()) 
		{
			throw new InvalidDataException("Property '"+property+"' has one or more invalid characters. Only a-z, A-Z, 0-9, _ and . is allowed.");
		}
		if (alias != null && !ALIAS_REGEXP.matcher(alias).matches()) 
		{
			throw new InvalidDataException("Alias '"+alias+"' has one or more invalid characters. Only a-z, A-Z and 0-9 is allowed.");
		}
		return new HqlIndexExpression(alias, property);
	}
	
	/**
		Create an elements expression for a property that is a collection
		of values. If no alias is given the property name is resolved 
		against the root entity of the query, otherwise it is resolved against the
		joined entity with the given alias.
		
		@param property The property name of the object which must set, map or other collection
		@return An expression representing the values collection of the property
		@throws InvalidDataException If both the alias and property parameters are null
			or if any of them contains invalid characters
	*/
	public static Expression elements(String alias, String property)
		throws InvalidDataException
	{
		if (property == null) throw new InvalidUseOfNullException("property");
		if (!PROPERTY_REGEXP.matcher(property).matches()) 
		{
			throw new InvalidDataException("Property '"+property+"' has one or more invalid characters. Only a-z, A-Z, 0-9, _ and . is allowed.");
		}
		if (alias != null && !ALIAS_REGEXP.matcher(alias).matches()) 
		{
			throw new InvalidDataException("Alias '"+alias+"' has one or more invalid characters. Only a-z, A-Z and 0-9 is allowed.");
		}
		return new HqlElementsExpression(alias, property);
	}

	/**
		Create a size expression for a property that is a collection
		of values. If no alias is given the property name is resolved 
		against the root entity of the query, otherwise it is resolved against the
		joined entity with the given alias.
		
		@param property The property name of the object which must set, map or other collection
		@return An expression representing the size of the collection
		@throws InvalidDataException If both the alias and property parameters are null
			or if any of them contains invalid characters
		@since 2.7
	*/
	public static Expression size(String alias, String property)
		throws InvalidDataException
	{
		if (property == null) throw new InvalidUseOfNullException("property");
		if (!PROPERTY_REGEXP.matcher(property).matches()) 
		{
			throw new InvalidDataException("Property '"+property+"' has one or more invalid characters. Only a-z, A-Z, 0-9, _ and . is allowed.");
		}
		if (alias != null && !ALIAS_REGEXP.matcher(alias).matches()) 
		{
			throw new InvalidDataException("Alias '"+alias+"' has one or more invalid characters. Only a-z, A-Z and 0-9 is allowed.");
		}
		return new HqlSizeExpression(alias, property);
	}

	
	/**
		Same as <code>innerJoin(null, propert, joinedAlias)</code>
		@see #innerJoin(String, String, String)
	*/
	public static Join innerJoin(String property, String joinedAlias)
		throws InvalidDataException
	{
		return innerJoin(null, property, joinedAlias, false);
	}

	/**
		Create an inner join query element. An inner join only selects an
		item if it has an associated item. The property must reference an
		association to another item or collection of values. See {@link
		#property(String, String)} for a code example.
		
		@param alias The alias to resolve the property against, or null to resolve the
			property against the root entity
		@param property The property name of the associated entity (required)
		@param joinedAlias The alias to give the joined entity (required)
		@return A join query element
		@throws InvalidDataException If the property or joined alias parameter
			are null or if any of the parameters contains invalid characters
	*/
	public static Join innerJoin(String alias, String property, String joinedAlias)
		throws InvalidDataException
	{
		return innerJoin(alias, property, joinedAlias, false);
	}

	/**
		Same as <code>innerJoin(null, propert, joinedAlias, null, false)</code>
		@see #innerJoin(String, String, String, Restriction, boolean)
	*/
	public static Join innerJoin(String alias, String property, String joinedAlias, boolean fetch)
		throws InvalidDataException
	{
		return innerJoin(alias, property, joinedAlias, null, fetch);
	}

	/**
		Create an inner join query element and optionally 
		prefetch the joined items in the same query. This is useful to 
		avoid subsequent queries to the database when you know that you are
		going to access the joined items in your code. If you are joining a
		collection the {@link Query#setFirstResult(int)} and {@link Query#setMaxResults(int)}
		shouldn't be used since the join will create multiple rows for the same
		root item.
		
		@param alias The alias to resolve the property against, or null to resolve the
			property against the root entity
		@param property The property name of the associated entity (required)
		@param joinedAlias The alias to give the joined entity (required)
		@param on Optional extra restriction that is used in the <code>ON</code> clause
			of the generated query
		@param fetch If the joined items should be prefetched or not
		@return A join query element
		@throws InvalidDataException If the property or joined alias parameter
			are null or if any of the parameters contains invalid characters
		@since 2.8
	*/
	public static Join innerJoin(String alias, String property, String joinedAlias, Restriction on, boolean fetch)
		throws InvalidDataException
	{
		if (property == null) throw new InvalidUseOfNullException("property");
		if (joinedAlias == null) throw new InvalidUseOfNullException("alias");
		if (!PROPERTY_REGEXP.matcher(property).matches()) 
		{
			throw new InvalidDataException("Property '"+property+"' has one or more invalid characters. Only a-z, A-Z, 0-9, _ and . is allowed.");
		}
		if (alias != null && !ALIAS_REGEXP.matcher(alias).matches()) 
		{
			throw new InvalidDataException("Alias '"+alias+"' has one or more invalid characters. Only a-z, A-Z and 0-9 is allowed.");
		}
		if (!ALIAS_REGEXP.matcher(joinedAlias).matches()) 
		{
			throw new InvalidDataException("Alias '"+joinedAlias+"' has one or more invalid characters. Only a-z, A-Z and 0-9 is allowed.");
		}
		return new HqlInnerJoin(alias, property, joinedAlias, on, fetch);
	}
	
	/**
		Same as <code>leftJoin(null, propert, joinedAlias, null, false)</code>
		@see #leftJoin(String, String, String, Restriction, boolean)
	*/
	public static Join leftJoin(String property, String joinedAlias)
		throws InvalidDataException
	{
		return leftJoin(null, property, joinedAlias, null, false);
	}

	/**
		Create a left join query element. A left join selects all items
		from the root entity even if they don't have an associated item. 
		The property must reference an association to another item or collection of values. 

<pre class="code">
<b>// Find all users not assigned to a role</b>
ItemQuery&lt;User&gt; query = User.getQuery();
query.join(Hql.leftJoin("roles", "rle"));
query.restrict(
  Restrictions.eq(
    Hql.alias("rle"),
    null
  )
);

<b>// Generated HQL</b>
SELECT usr 
FROM UserData usr
LEFT JOIN usr.roles rle
WHERE rle IS NULL
</pre>

		<p>
		The <code>fetch</code> parameter is used to fetch the joined data in the
		same query. This is useful to 
		avoid subsequent queries to the database when you know that you are
		going to access the joined items in your code. If you are joining a
		collection the {@link Query#setFirstResult(int)} and {@link Query#setMaxResults(int)}
		shouldn't be used since the join will create multiple rows for the same
		root item.

		@param alias The alias to resolve the property against, or null to resolve the
			property against the root entity
		@param property The property name of the associated entity (required)
		@param joinedAlias The alias to give the joined entity (required)
		@param on Optional extra restriction that is used in the <code>ON</code> clause
			of the generated query
		@param fetch If the joined items should be prefetched or not
		@return A join query element
		@throws InvalidDataException If the property or joined alias parameter
			are null or if any of the parameters contains invalid characters
	*/
	public static Join leftJoin(String alias, String property, String joinedAlias, Restriction on, boolean fetch)
		throws InvalidDataException
	{
		if (property == null) throw new InvalidUseOfNullException("property");
		if (joinedAlias == null) throw new InvalidUseOfNullException("alias");
		if (!PROPERTY_REGEXP.matcher(property).matches()) 
		{
			throw new InvalidDataException("Property '"+property+"' has one or more invalid characters. Only a-z, A-Z, 0-9, _ and . is allowed.");
		}
		if (alias != null && !ALIAS_REGEXP.matcher(alias).matches()) 
		{
			throw new InvalidDataException("Alias '"+alias+"' has one or more invalid characters. Only a-z, A-Z and 0-9 is allowed.");
		}
		if (!ALIAS_REGEXP.matcher(joinedAlias).matches()) 
		{
			throw new InvalidDataException("Alias '"+joinedAlias+"' has one or more invalid characters. Only a-z, A-Z and 0-9 is allowed.");
		}
		return new HqlLeftJoin(alias, property, joinedAlias, on, fetch);
	}

	/**
		Same as <code>rightJoin(null, propert, joinedAlias, null)</code>
		@see #rightJoin(String, String, String, Restriction)
	*/
	public static Join rightJoin(String property, String joinedAlias)
		throws InvalidDataException
	{
		return rightJoin(null, property, joinedAlias, null);
	}

	/**
		Create a right join query element. A right join selects all items
		from the joined association even if they don't have an associated item in 
		the root entity. The property must reference an association to another 
		item or collection of values. 

		@param alias The alias to resolve the property against, or null to resolve the
			property against the root entity
		@param property The property name of the associated entity (required)
		@param joinedAlias The alias to give the joined entity (required)
		@param on Optional extra restriction that is used in the <code>ON</code> clause
			of the generated query
		@return A join query element
		@throws InvalidDataException If the property or joined alias parameter
			are null or if any of the parameters contains invalid characters
	*/
	public static Join rightJoin(String alias, String property, String joinedAlias, Restriction on)
		throws InvalidDataException
	{
		if (property == null) throw new InvalidUseOfNullException("property");
		if (joinedAlias == null) throw new InvalidUseOfNullException("alias");
		if (!PROPERTY_REGEXP.matcher(property).matches()) 
		{
			throw new InvalidDataException("Property '"+property+"' has one or more invalid characters. Only a-z, A-Z, 0-9, _ and . is allowed.");
		}
		if (alias != null && !ALIAS_REGEXP.matcher(alias).matches()) 
		{
			throw new InvalidDataException("Alias '"+alias+"' has one or more invalid characters. Only a-z, A-Z and 0-9 is allowed.");
		}
		if (!ALIAS_REGEXP.matcher(joinedAlias).matches()) 
		{
			throw new InvalidDataException("Alias '"+joinedAlias+"' has one or more invalid characters. Only a-z, A-Z and 0-9 is allowed.");
		}
		return new HqlRightJoin(alias, property, joinedAlias, on);
	}
	
	/**
	 	@param restrictionString A restriction as a string. 
	 		Null or empty string is not allowed.
	 	@param prefix Prefix used in the restriction string. 
	 		The prefix will be replaced in the returned object by the query's root-alias. 
	 		Null is allowed.
	 	@return a Restriction to use in a query.
	 	@since 2.5
	 */
	public static Restriction restriction(String restrictionString, String prefix)
	{		
		if (restrictionString == null) throw new InvalidUseOfNullException("restrictionString");
		if (restrictionString.length() == 0) throw new InvalidDataException("Restriction string can not be empty.");
		return new HqlExpression(restrictionString, prefix);
	}
	
	/**
	 	@param exprString An expression as a string.
	 		Null or empty string is not allowed.
	 	@param prefix Prefix used in the expression string.
	 		The prefix will be replaced in the returned object by the query's root-alias.
	 		Null is allowed.
	 	@return an Object implementing {@link net.sf.basedb.core.query.Expression}.
	 	@since 2.5
	 */
	public static Expression expression(String exprString, String prefix)
	{
		if (exprString == null) throw new InvalidUseOfNullException("exprString");
		if (exprString.length() == 0) throw new InvalidDataException("Expression string can not be empty");
		return new HqlExpression(exprString, prefix);
	}
	
	public static Restriction isPartOf(String alias, String property, ReporterList reporterList)
	{
		if (reporterList == null) throw new InvalidUseOfNullException("reporterList");
		if (property == null) throw new InvalidUseOfNullException("property");
		if (!PROPERTY_REGEXP.matcher(property).matches()) 
		{
			throw new InvalidDataException("Property '"+property+"' has one or more invalid characters. Only a-z, A-Z, 0-9, _ and . is allowed.");
		}
		if (alias != null && !ALIAS_REGEXP.matcher(alias).matches()) 
		{
			throw new InvalidDataException("Alias '"+alias+"' has one or more invalid characters. Only a-z, A-Z and 0-9 is allowed.");
		}
		return Restrictions.neq(
			new ReporterListExpression(reporterList, alias, property, "reporterId", JoinType.INNER), 
			null
		);
	}
	
	public static Restriction isNotPartOf(String alias, String property, ReporterList reporterList)
	{
		if (reporterList == null) throw new InvalidUseOfNullException("reporterList");
		if (alias != null && !ALIAS_REGEXP.matcher(alias).matches()) 
		{
			throw new InvalidDataException("Alias '"+alias+"' has one or more invalid characters. Only a-z, A-Z and 0-9 is allowed.");
		}
		if (property == null) throw new InvalidUseOfNullException("property");
		if (!PROPERTY_REGEXP.matcher(property).matches()) 
		{
			throw new InvalidDataException("Property '"+property+"' has one or more invalid characters. Only a-z, A-Z, 0-9, _ and . is allowed.");
		}
		return Restrictions.eq(
			new ReporterListExpression(reporterList, alias, property, "reporterId", JoinType.LEFT),
			null
		);
	}
	
	/**
		A special restriction that works on {@link Shareable} items. 
		The restriction is used to find items that have or have not
		been shared to users/groups and/or projects that meet some
		specific criteria.
		
		@param sharedTo TRUE to finds items that have been shared to at
			least one of the matching users/groups/projects, FALSE to find
			items that are not shared to any of the matching users/groups/projects.
			If all three restrictions are null, a value of TRUE will find items
			that are not shared and FALSE will find items that are shared to
			anyone.
		@param users A restriction to apply on the subquery that is matching users, 
			or null to not match against users
		@param groups A restriction to apply on the subquery that is matching groups, 
			or null to not match against groups
		@param projects A restriction to apply on the subquery that is matching projects, 
			or null to not match against projects
		@return A restriction
		@since 2.15
	*/
	public static Restriction sharedTo(boolean sharedTo, Restriction users, Restriction groups, Restriction projects)
	{
		return new SharedToRestriction(sharedTo, users, groups, projects);
	}
}
