/**
	$Id: CdfValidator.java 5623 2011-05-06 12:09:13Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.affymetrix;

import affymetrix.fusion.cdf.FusionCDFData;
import net.sf.basedb.core.ArrayDesign;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.File;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.core.InvalidUseOfNullException;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.Reporter;
import net.sf.basedb.core.ReporterBatcher;
import net.sf.basedb.core.ReporterType;
import net.sf.basedb.core.data.ReporterData;
import net.sf.basedb.core.signal.ThreadSignalHandler;

/**
	Helper class for working with Affymetrix CDF files.

	@author Nicklas
	@since 3.0
	@base.modified $Date: 2011-05-06 14:09:13 +0200 (fr, 06 maj 2011) $
*/
public class CdfValidator
{

	/**
		Create a new instance.
	*/
	public CdfValidator()
	{}

	/**
		Load a CDF file using the Affymetric Fusion SDK. This method
		checks that the file exists, and read all headers.
		
		@param cdfFile The file to load
		@return A FusionCDFData object representing the CDF file
		@throws ItemNotFoundException If the actual file is not on the server
		@throws InvalidDataException If the file is not a CDF file
	*/
	public FusionCDFData loadCdfFile(File cdfFile)
		throws ItemNotFoundException, InvalidDataException
	{
		if (cdfFile == null) throw new InvalidUseOfNullException("cdfFile");
		if (!cdfFile.getLocation().isDownloadable())
		{
			throw new ItemNotFoundException("Can't download file '" + 
				cdfFile.getName() + "'; location=" + cdfFile.getLocation());
		}
		FusionCDFData cdf = new FusionCDFData();
		cdf.setFileName(cdfFile.getName());
		cdf.setInputStream(cdfFile.getDownloadStream(0));
		if (!cdf.readHeader())
		{
			throw new InvalidDataException("Could not read CDF file '" +
					cdfFile.getPath()+"': " + cdf.getError());
		}
		return cdf;
	}

	/**
		Copy metadata from the CDF file to the array design.
		This sets the number of features to the number of 
		probesets.
		
		@param cdf The CDF file
		@param design The array design
	*/
	public void copyMetadata(FusionCDFData cdf, ArrayDesign design)
	{
		design.setNumFileFeatures(cdf.getHeader().getNumProbeSets());
	}
	
	/**
		Reset metadata on the given array design. This set the
		number of features to 0.
		@param design The array design
	*/
	public void resetMetadata(ArrayDesign design)
	{
		design.setNumFileFeatures(0);
	}
	
	/**
		Verify if probesets in a CDF file exists as reporters and
		optionally create those that doesn't exists. This method will clear
		the CDF file, read it, and then clear it again. The calling code is 
		responsible for committing the transaction.
		
		@param cdf The CDF file
		@param batcher A reporter batcher used to check and insert reporters
		@param create TRUE to create non-existing reporters, FALSE to throw
			an ItemNotFoundException
		@param type The reporter type to assign to new reporters (if create is true)
		@param progress An optional progress reporter 
		@return The number of created reporter
		@throws ItemNotFoundException If create is FALSE and a reporter isn't found
		@throws BaseException If there is another error
		@since 2.4
	*/
	public int checkIfReportersExists(FusionCDFData cdf, ReporterBatcher batcher,
		boolean create, ReporterType type, ProgressReporter progress)
		throws ItemNotFoundException, BaseException
	{
		if (batcher == null) throw new InvalidUseOfNullException("batcher");
		if (cdf == null) throw new InvalidUseOfNullException("cdf");
	
		final boolean hasType = type != null;
		final boolean hasProgress = progress != null;
		int numCreated = 0;
		
		if (hasProgress) progress.display(5, "Reading CDF file...");
		if (!cdf.read())
		{
			throw new BaseException("Error reading CDF file: " + cdf.getError());
		}
		int numProbesets = cdf.getHeader().getNumProbeSets();
		int progressInterval = numProbesets / 50;
		if (progressInterval < 10) progressInterval = 10;
		
		int index = 0;
		while (index < numProbesets)
		{
			ThreadSignalHandler.checkInterrupted();
			String probesetId = cdf.getProbeSetName(index);
			if (!batcher.exists(probesetId, true, true))
			{
				if (create)
				{
					ReporterData reporter = Reporter.getNew(probesetId);
					if (hasType) Reporter.setReporterType(reporter, type);
					reporter.setName(probesetId);
					batcher.insert(reporter);
					++numCreated;
				}
				else
				{
					throw new ItemNotFoundException("Reporter[externalId=" + probesetId+"]");
				}
			}
			++index;
			if (hasProgress && (index % progressInterval == 0))
			{
				int percent = 5 + (int)(90L * (long)index / numProbesets);
				int remain = numProbesets - index;
				String created = create ? "; " + numCreated + " created" : "";
				String msg = "Checking probsets (" + index + " done" + created + "; " + remain + " remaining)...";
				progress.display(percent, msg);
			}
		}
		batcher.flush();
		cdf.clear();
		if (hasProgress) progress.display(100, "Checked: " + numProbesets + "; created: " + numCreated);
		return numCreated;
	}
	
}
