/**
	$Id$
	
	Copyright (C) 2015 Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.loader;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemList;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultIterator;
import net.sf.basedb.core.Listable;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.node.ChildNodeDirection;
import net.sf.basedb.util.overview.node.ItemListMemberNameGenerator;
import net.sf.basedb.util.overview.node.NameableNameGenerator;
import net.sf.basedb.util.overview.node.NodeFactory;
import net.sf.basedb.util.overview.validator.NodeValidator;


/**
	Node loader implementation for item lists. Item
	lists can only be loaded as root nodes. The forward-loading
	direction goes to the item that is member of the list.
	
	@author nicklas
	@since 3.5
*/
public class ItemListLoader 
	extends BasicItemNodeLoader<ItemList> 
{
	public ItemListLoader()
	{
		super(Item.ITEMLIST, ALLOW_ROOT_NODE, 
				new NameableNameGenerator<ItemList>("list", "List"));
	}
	
	/*
		From the AbstractNodeLoader class
		----------------------------------
	*/	
	/**
		Loads property nodes of a item list. 
		<ul>
		<li>Annotations: {@link AnnotationLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Any-to-any: {@link AnyToAnyLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		</ul>
	*/
	@Override
	protected void loadPropertyChildNodes(DbControl dc, OverviewContext context, Node listNode)
	{
		getNodeLoader(context, Item.ANNOTATION).createPropertyNode(dc, context, listNode);
		getNodeLoader(context, Item.ANYTOANY).createPropertyNode(dc, context, listNode);
	}

	/**
		Loads the list members.
	*/
	@Override
	protected void loadForwardChildNodes(DbControl dc, OverviewContext context, Node listNode)
	{
		ItemList list = (ItemList)listNode.getItem(dc);
		getNodeLoader(context, list.getMemberType()).createForwardNode(dc, context, listNode);
	}
	// ---------------------------------------

	/**
		Utiltity method to be used by other  loaders to load members of an item list.
		The loader should be one that loads {@link Listable} items.
		The member nodes are created as child nodes to the list node without any 
		intermediate folder node.
		
		@param loader The loader implementation to load member items
		@param listNode The parent item list node
	*/
	@SuppressWarnings("unchecked")
	public static <T extends BasicItem & Listable> void loadMemberNodes(BasicItemNodeLoader<T> loader, DbControl dc, OverviewContext context, Node listNode)
	{
		NodeValidator<T> validator = loader.getNodeValidator(context);
		NodeFactory<T> factory = loader.getNodeFactory(dc, context, validator, new ItemListMemberNameGenerator());
		ItemList list = (ItemList)listNode.getItem(dc);
		
		// Find all members of the list
		ItemQuery<T> query = (ItemQuery<T>)context.initQuery(list.getMembers(), "name");
		
		ItemResultIterator<T> it = query.iterate(dc);
		while (it.hasNext())
		{
			T item = it.next();
			loader.createItemNode(factory, item, item, false, listNode, ChildNodeDirection.ALL);
		}
	}

}
