/**
	$Id$

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.plugins.batchimport;

import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.regex.Pattern;

import net.sf.basedb.core.AnnotationType;
import net.sf.basedb.core.AnnotationTypeCategory;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.PluginParameter;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.Unit;
import net.sf.basedb.core.plugin.GuiContext;
import net.sf.basedb.core.plugin.Request;
import net.sf.basedb.core.plugin.Response;
import net.sf.basedb.util.Values;
import net.sf.basedb.util.parser.FlatFileParser;
import net.sf.basedb.util.parser.FlatFileParser.Data;
import net.sf.basedb.util.parser.Mapper;

/**
	Plug-in for importing annotation type items in a batch. The plug-in can create new 
	items and updated existing items.

	@author nicklas
	@since 3.8
*/
public class AnnotationTypeImporter
	extends AbstractItemImporter<AnnotationType>
{

	private static final Set<GuiContext> guiContexts =  
		Collections.singleton(new GuiContext(Item.ANNOTATIONTYPE, GuiContext.Type.LIST));
	
	
	/**
		Column mapping parameter for the annotation value type
	*/
	private static final PluginParameter<String> valueTypeColumnMapping = 
		new PluginParameter<String>(
			"valueTypeColumnMapping",
			"Value type",
			"Mapping that picks value type (eg. INT, FLOAT, STRING, etc.) from the data columns. " +
			"Required when creating new annotation types and ignored for existing since it is " +
			"not possible to change the value type. " +
			"Example: \\Value type\\",
			optionalColumnMapping
		);

	/**
		Column mapping parameter for assigning annotation types to categories.
	*/
	private static final PluginParameter<String> categoryColumnMapping = 
		new PluginParameter<String>(
			"categoryColumnMapping",
			"Category",
			"Mapping that picks name or id of an annotation type category from the data columns. " +
			"More than one category should be specified on multiple lines. "+
			"Example: \\Category\\",
			optionalColumnMapping
		);

	/**
		Column mapping parameter for annotation multiplicty.
	*/
	private static final PluginParameter<String> multiplicityColumnMapping = 
		new PluginParameter<String>(
			"multiplicityColumnMapping",
			"Multiplicity",
			"Mapping that picks the multiplicity from the data columns. " +
			"Example: \\Multiplicity\\",
			optionalColumnMapping
		);

	/**
		Column mapping parameter for "Default value".
	*/
	private static final PluginParameter<String> defaultValueColumnMapping = 
		new PluginParameter<String>(
			"defaultValueColumnMapping",
			"Default value",
			"Mapping that picks the 'Default value' from the data columns. " +
			"Example: \\Default value\\",
			optionalColumnMapping
		);

	/**
		Column mapping parameter for "Required for MIAME" flag.
	*/
	private static final PluginParameter<String> miameColumnMapping = 
		new PluginParameter<String>(
			"miameColumnMapping",
			"Requried for MIAME",
			"Mapping that picks the 'Required for MIAME' flag from the data columns. " +
			"Example: \\Required for MIAME\\",
			optionalColumnMapping
		);

	/**
		Column mapping parameter for "Protocol parameter" flag.
	*/
	private static final PluginParameter<String> protocolParameterColumnMapping = 
		new PluginParameter<String>(
			"protocolParameterColumnMapping",
			"Protocol parameter",
			"Mapping that picks the 'Protocol parameter' flag from the data columns. " +
			"Example: \\Protocol parameter\\",
			optionalColumnMapping
		);
	
	/**
		Column mapping parameter for "Disabled history log" flag.
	*/
	private static final PluginParameter<String> disableChangeLogColumnMapping = 
		new PluginParameter<String>(
			"disableChangeLogColumnMapping",
			"Disable history log",
			"Mapping that picks the 'Disable history log' flag from the data columns. " +
			"Example: \\Disable history log\\",
			optionalColumnMapping
		);
	
	/**
		Column mapping parameter for "Disable inheritance" flag.
	*/
	private static final PluginParameter<String> disableInheritanceColumnMapping = 
		new PluginParameter<String>(
			"disableInheritanceColumnMapping",
			"Disable inheritance",
			"Mapping that picks the 'Disable inheritance' flag from the data columns. " +
			"Example: \\Disable inheritance\\",
			optionalColumnMapping
		);
	
	/**
		Column mapping parameter for "Item types".
	*/
	private static final PluginParameter<String> itemTypesColumnMapping = 
		new PluginParameter<String>(
			"itemTypesColumnMapping",
			"Item types",
			"Mapping that picks the 'Item types' setting from the data columns. The "+
			"values should be a comma-separated list of item type names (eg. SAMPLE,EXTRACT)." +
			"Example: \\Item types\\",
			optionalColumnMapping
		);

	/**
		Column mapping parameter for "Default unit".
	*/
	private static final PluginParameter<String> defaultUnitColumnMapping = 
		new PluginParameter<String>(
			"defaultUnitColumnMapping",
			"Default unit",
			"Mapping that picks the 'Default unit' setting from the data columns. The value " +
			"can be either the descriptive name of the unit or one of the symbols " +
			"(eg. Microliter or µl, Meter per second or m/s). The unit column is ignored " +
			"for non-numerical annotation types. " +
			"Example: \\Default unit\\",
			optionalColumnMapping
		);
	
	/**
		Column mapping parameter for "Usable unit".
	*/
	private static final PluginParameter<String> usableUnitColumnMapping = 
		new PluginParameter<String>(
			"usableUnitColumnMapping",
			"Usable unit",
			"Mapping that picks the 'Usable unit' setting from the data columns. The value " +
			"can be either the descriptive name of the unit or one of the symbols " +
			"(eg. Microliter or µl, Meter per second or m/s). This column is ignored " +
			"for non-numerical annotation types. More than one unit should be specified on multiple lines. " +
			"Example: \\Usable unit\\",
			optionalColumnMapping
		);
	
	/**
		Column mapping parameter for "Enumeration".
	*/
	private static final PluginParameter<String> enumerationColumnMapping = 
		new PluginParameter<String>(
			"enumerationColumnMapping",
			"Enumeration",
			"Mapping that picks the 'Enumeration' setting from the data columns. Ignored " +
			"for non-enumerable annotation types. The value in this column is split by the "+
			"'Enumeration splitter' parameter (default is comma) to produce the list of allowed values. "+
			"Example: \\Enumeration\\",
			optionalColumnMapping
		);

	/**
		Regular expression splitter parameter for enumerations.
	*/
	private static final PluginParameter<String> enumerationSplitterRegexpParameter = 
		new PluginParameter<String>(
			"enumerationSplitterRegexp",
			"Enumeration splitter",
			"A regular expression that splits an enumeration column value into multiple values. "+
			"If not specified, comma is used.",
			optionalColumnMapping
		);
	
	
	/**
		Column mapping parameter for "Min value".
	*/
	private static final PluginParameter<String> minValueColumnMapping = 
		new PluginParameter<String>(
			"minValueColumnMapping",
			"Min value",
			"Mapping that picks the 'Min value' setting from the data columns. Ignored " +
			"for non-numerical annotation types. " +
			"Example: \\Min value\\",
			optionalColumnMapping
		);
	
	/**
		Column mapping parameter for "Max value".
	*/
	private static final PluginParameter<String> maxValueColumnMapping = 
		new PluginParameter<String>(
			"maxValueColumnMapping",
			"Max value",
			"Mapping that picks the 'Max value' setting from the data columns. Ignored " +
			"for non-numerical annotation types. " +
			"Example: \\Max value\\",
			optionalColumnMapping
		);
	
	public AnnotationTypeImporter()
	{}

	/*
		From the InteractivePlugin interface
		------------------------------------
	*/
	@Override
	public Set<GuiContext> getGuiContexts()
	{
		return guiContexts;
	}
	/**
		Override this method so that we can make some extra parameter checks:
		* A mapping for 'Value type' must be specified when 'create' or 'create+update' mode is used
	 */
	@Override
	public void configure(GuiContext context, Request request, Response response)
	{
		String command = request.getCommand();
		try
		{
			if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN) || command.equals(Request.COMMAND_CONFIGURE_JOB))
			{
				String importMode = (String)request.getParameterValue("mode");
				if (importMode != null && importMode.contains("create"))
				{
					String valueTypeMapping = (String)request.getParameterValue(valueTypeColumnMapping.getName());
					if (valueTypeMapping == null)
					{
						response.setError("A colum mapping for 'Value type' is required when using 'create' mode.", null);
						return;
					}
				}
			}
			super.configure(context, request, response);
		}
		catch (Throwable ex)
		{
			response.setError(ex.getMessage(), Arrays.asList(ex));
		}
	}
	// ------------------------------------

	/*
		From the AbstractItemImporter class
		------------------------------------
	*/
	
	private Mapper nameMapper;
	private Mapper externalIdMapper;
	private Mapper descriptionMapper;
	private Mapper valueTypeMapper;
	private Mapper categoryMapper;
	private Mapper registeredMapper;
	private Mapper multiplicityMapper;
	private Mapper defaultValueMapper;
	private Mapper miameMapper;
	private Mapper protocolParameterMapper;
	private Mapper disableChangeLogMapper;
	private Mapper disableInheritanceMapper;
	private Mapper itemTypesMapper;
	private Mapper defaultUnitMapper;
	private Mapper usableUnitMapper;
	private Mapper enumerationMapper;
	private Mapper minValueMapper;
	private Mapper maxValueMapper;
	
	private NumberFormat numberFormat;
	private Pattern enumerationSplitter;
	
	/**
		Useable methods are:
		<ul>
		<li>{@link PropertyIdMethod#NAME}
		<li>{@link PropertyIdMethod#EXTERNAL_ID}
		<li>{@link InternalIdMethod#INTERNAL_ID}
		<li>{@link FallbackIdMethod#NAME_OR_ID}
		</ul>
	*/
	@Override
	protected IdMethod[] getIdMethods()
	{
		return new IdMethod[] { PropertyIdMethod.NAME, PropertyIdMethod.EXTERNAL_ID,
				InternalIdMethod.INTERNAL_ID, FallbackIdMethod.NAME_OR_ID};
	}
	
	@Override
	protected void addMoreParsingParameters(List<PluginParameter<?>> parameters) 
	{
		parameters.add(enumerationSplitterRegexpParameter);
	}

	/**
		Adds column mappings for name and description.
	*/
	@Override
	protected void addMoreColumnMappingParameters(List<PluginParameter<?>> parameters)
	{
		parameters.add(internalIdColumnMapping);
		parameters.add(nameColumnMapping);
		parameters.add(valueTypeColumnMapping);
		parameters.add(externalIdColumnMapping);
		parameters.add(descriptionColumnMapping);
		parameters.add(categoryColumnMapping);
		
		parameters.add(registeredColumnMapping);
		parameters.add(multiplicityColumnMapping);
		parameters.add(defaultValueColumnMapping);
		parameters.add(miameColumnMapping);
		parameters.add(protocolParameterColumnMapping);
		parameters.add(disableChangeLogColumnMapping);
		parameters.add(disableInheritanceColumnMapping);

		parameters.add(itemTypesColumnMapping);
		parameters.add(defaultUnitColumnMapping);
		parameters.add(usableUnitColumnMapping);
		
		parameters.add(enumerationColumnMapping);
		parameters.add(minValueColumnMapping);
		parameters.add(maxValueColumnMapping);
		
		parameters.add(permissionTemplateColumnMapping);
	}
	/**
		Calls {@link AnnotationType#getQuery(Item)}.
	*/
	@Override
	protected ItemQuery<AnnotationType> createItemQuery()
	{
		return AnnotationType.getQuery(null);
	}
	@Override
	protected void createColumnMappers(FlatFileParser ffp, boolean cropStrings)
	{
		nameMapper = getMapper(ffp, (String)job.getValue("nameColumnMapping"), 
				cropStrings ? AnnotationType.MAX_NAME_LENGTH : null, null);
		valueTypeMapper = getMapper(ffp, (String)job.getValue("valueTypeColumnMapping"), null, null);
		externalIdMapper = getMapper(ffp, (String)job.getValue("externalIdColumnMapping"), 
				cropStrings ? AnnotationType.MAX_EXTERNAL_ID_LENGTH : null, null);
		categoryMapper = getMapper(ffp, (String)job.getValue("categoryColumnMapping"), null, null);
		descriptionMapper = getMapper(ffp, (String)job.getValue("descriptionColumnMapping"), 
				cropStrings ? AnnotationType.MAX_DESCRIPTION_LENGTH : null, null);
		registeredMapper = getMapper(ffp, (String)job.getValue("registeredColumnMapping"), null, null);
		
		multiplicityMapper = getMapper(ffp, (String)job.getValue("multiplicityColumnMapping"), null, null);
		defaultValueMapper = getMapper(ffp, (String)job.getValue("defaultValueColumnMapping"), 
				cropStrings ? AnnotationType.MAX_DEFAULT_VALUE_LENGTH : null, null);
		miameMapper = getMapper(ffp, (String)job.getValue("miameColumnMapping"), null, null);
		protocolParameterMapper = getMapper(ffp, (String)job.getValue("protocolParameterColumnMapping"), null, null);
		disableChangeLogMapper = getMapper(ffp, (String)job.getValue("disableChangeLogColumnMapping"), null, null);
		disableInheritanceMapper = getMapper(ffp, (String)job.getValue("disableInheritanceColumnMapping"), null, null);
		itemTypesMapper = getMapper(ffp, (String)job.getValue("itemTypesColumnMapping"), null, null);
		defaultUnitMapper = getMapper(ffp, (String)job.getValue("defaultUnitColumnMapping"), null, null);
		usableUnitMapper = getMapper(ffp, (String)job.getValue("usableUnitColumnMapping"), null, null);

		enumerationMapper = getMapper(ffp, (String)job.getValue("enumerationColumnMapping"), null, null);
		if (enumerationMapper != null)
		{
			enumerationSplitter = getPattern("enumerationSplitterRegexp");
			if (enumerationSplitter == null) enumerationSplitter = Pattern.compile(",");
		}
		minValueMapper = getMapper(ffp, (String)job.getValue("minValueColumnMapping"), null, null);
		maxValueMapper = getMapper(ffp, (String)job.getValue("maxValueColumnMapping"), null, null);
		
		numberFormat = ffp.getDefaultNumberFormat();
	}
	@Override
	protected AnnotationType createItem(DbControl dc, FlatFileParser.Data data)
	{
		Type valueType = Type.valueOf(valueTypeMapper.getValue(data));
		AnnotationType at = AnnotationType.getNew(dc, valueType);
		if (registeredMapper != null) 
		{
			at.setEntryDate(parseDate(registeredMapper.getValue(data)));
		}
		updateItem(dc, at, data);
		return at;
	}
	@Override
	protected void updateItem(DbControl dc, AnnotationType at, FlatFileParser.Data data)
	{
		if (nameMapper != null) at.setName(nameMapper.getValue(data));
		if (descriptionMapper != null) at.setDescription(descriptionMapper.getValue(data));
		if (externalIdMapper != null) at.setExternalId(externalIdMapper.getValue(data));
		if (multiplicityMapper != null) at.setMultiplicity(multiplicityMapper.getInt(data));
		if (defaultValueMapper != null) at.setDefaultValue(defaultValueMapper.getValue(data));
		if (miameMapper != null) at.setRequiredForMiame(Values.getBoolean(miameMapper.getValue(data)));
		if (protocolParameterMapper != null) at.setProtocolParameter(Values.getBoolean(protocolParameterMapper.getValue(data)));
		if (disableChangeLogMapper != null) at.setDisableLogOfValues(Values.getBoolean(disableChangeLogMapper.getValue(data)));
		if (disableInheritanceMapper != null) at.setDisableInheritance(Values.getBoolean(disableInheritanceMapper.getValue(data)));
		if (itemTypesMapper != null)
		{
			String itemTypes = itemTypesMapper.getValue(data);
			Set<Item> items = new HashSet<Item>();
			if (itemTypes != null)
			{
				for (String item : itemTypes.split(","))
				{
					items.add(Item.valueOf(item.trim()));
				}
			}
			at.setItemTypes(items);
		}
		if (defaultUnitMapper != null && at.getValueType().isNumerical())
		{
			String nameOrSymbolOrId = defaultUnitMapper.getValue(data);
			Unit unit = findUnit(dc, FallbackIdMethod.NAME_OR_UNITSYMBOL_OR_ID, nameOrSymbolOrId);
			if (unit != null) 
			{
				at.setDefaultUnit(unit);
				if (usableUnitMapper != null) at.addUsableUnit(unit);
			}
		}
		
		Type valueType = at.getValueType();
		if (enumerationMapper != null && valueType.canEnumerate())
		{
			String enumeration = enumerationMapper.getValue(data);
			at.setEnumeration(enumeration != null);
			if (enumeration != null)
			{
				at.setDisplayAsList(false);
				List<Object> values = new ArrayList<Object>();
				for (String value : enumerationSplitter.split(enumeration))
				{
					Object v = valueType.parseString(value, numberFormat, true);
					if (v != null) values.add(v);
				}
				at.setValues(values);
			}
		}
		if (minValueMapper != null && valueType.isNumerical())
		{
			String minValue = minValueMapper.getValue(data);
			if (valueType == Type.INT || valueType == Type.LONG)
			{
				
				at.setMinValueLong((Long)Type.LONG.parseString(minValue, numberFormat, false));
			}
			else
			{
				at.setMinValueDouble((Double)Type.DOUBLE.parseString(minValue, numberFormat, false));
			}
		}
		if (maxValueMapper != null && valueType.isNumerical())
		{
			String maxValue = maxValueMapper.getValue(data);
			if (valueType == Type.INT || valueType == Type.LONG)
			{
				at.setMaxValueLong((Long)Type.LONG.parseString(maxValue, numberFormat, false));
			}
			else
			{
				at.setMaxValueDouble((Double)Type.DOUBLE.parseString(maxValue, numberFormat, false));
			}
		}
		
		updateMultiLineItem(dc, at, data, 0);
	}
	// ------------------------------------

	@Override
	protected void updateMultiLineItem(DbControl dc, AnnotationType at, Data data, int multiLineNum) 
	{
		if (categoryMapper != null)
		{
			String nameOrId = categoryMapper.getValue(data);
			AnnotationTypeCategory category = findAnnotationTypeCategory(dc, FallbackIdMethod.NAME_OR_ID, nameOrId);
			if (category != null) at.addCategory(category);
		}
		if (usableUnitMapper != null && at.supportUnits())
		{
			String nameOrSymbolOrId = usableUnitMapper.getValue(data);
			Unit unit = findUnit(dc, FallbackIdMethod.NAME_OR_UNITSYMBOL_OR_ID, nameOrSymbolOrId);
			if (unit != null) at.addUsableUnit(unit);
		}
	}
}
